# coding: utf-8
import math
import numpy as np


oh = 0.09572
hangle = 104.52 * math.pi / 180 / 2
mass=18
ohz = oh * math.cos(hangle)
ohy = oh * math.sin(hangle)
oz  = -ohz*2/mass


sites = np.array([[0, 0,oz],
                  [0, ohy,ohz+oz],
                  [0,-ohy,ohz+oz]]) # nm, OHHM

labels = ["O","H","H"]
name = "SOL"



import sys
import logging
import random
import itertools as it
import logging
from math import sin, cos, pi
from collections import Iterable, defaultdict

import numpy as np

from genice.importer import safe_import
from genice import pairlist as pl
from genice import digraph as dg
from genice import rigid
from genice.cells import rel_wrap, Cell

#for alkyl groups (Experimental)
from genice import alkyl

def load_numbers(v):
    if type(v) is str:
        return np.fromstring(v, sep=" ")
    elif type(v) is list:
        return np.array(v)
    else:
        return v


def complement(v):
    # assume missing vectors
    assert len(v) > 1
    if len(v) == 3:
        return [-(v[0]+v[1]+v[2])]
    elif len(v) == 2:
        y = v[1] - v[0]
        y /= np.linalg.norm(y)
        z = v[1] + v[0]
        z /= np.linalg.norm(z)
        x = np.cross(y,z)
        v2 = ( x*8.0**0.5 - z)/3.0
        v3 = (-x*8.0**0.5 - z)/3.0
        return v2,v3
    return []
        
        
        

def orientations(coord, graph, cell):
    """
    Does not work when two OHs are colinear
    """
    logger = logging.getLogger()
    rotmatrices = []
    for node in range(graph.number_of_nodes()):
        if node in graph.ignores:
            # for dopants; do not rotate
            rotmat = np.identity(3)
        else:
            vsucc = [cell.rel2abs(rel_wrap(coord[x] - coord[node])) for x in graph.successors(node)]
            if len(vsucc) < 2: #TSL
                vpred = [cell.rel2abs(rel_wrap(coord[x] - coord[node])) for x in graph.predecessors(node)]
                vsucc = [x / np.linalg.norm(x) for x in vsucc]
                vpred = [x / np.linalg.norm(x) for x in vpred]
                vcomp = complement(vpred+vsucc)
                logger.debug("Node {0} vcomp {1}".format(node,vcomp))    
                vsucc = (vsucc+vcomp)[:2]
            logger.debug("Node {0} vsucc {1}".format(node,vsucc))
            assert 2<=len(vsucc), "Probably a wrong ice network."
            y = vsucc[1] - vsucc[0]
            y /= np.linalg.norm(y)
            z = (vsucc[0] + vsucc[1]) / 2
            z /= np.linalg.norm(z)
            x = np.cross(y, z)
            rotmat = np.vstack([x, y, z])
            
        rotmatrices.append(rotmat)
    return rotmatrices


def arrange_atoms(coord, cell, rotmatrices, intra, labels, name, ignores=set()):
    atoms = []
    if len(intra) == 0:
        return atoms
    for order, pos in enumerate(coord):
        if order in ignores:
            continue
        abscom = cell.rel2abs(pos)  # relative to absolute
        rotated = np.dot(intra, rotmatrices[order])
        for i in range(len(labels)):
            atoms.append([i, name, labels[i], rotated[i, :] + abscom, order])
    return atoms


def shortest_distance(coord, cell, pairs=None):
    dmin = 1e99
    if pairs is None:
        iter = it.combinations(coord, 2)
    else:
        iter = [(coord[i], coord[j]) for i, j in pairs]
    for c1, c2 in iter:
        r = cell.rel2abs(rel_wrap(c1 - c2))
        rr = np.dot(r, r)
        if rr < dmin:
            dmin = rr
    return dmin**0.5


def flatten(item):
    """Yield items from any nested iterable; see REF."""
    if type(item) is str:
        yield item
    elif isinstance(item, Iterable):
        for x in item:
            yield from flatten(x)
    else:
        yield item


def replicate_groups(groups, waters, cagepos, rep):
    """
    This is not that easy.
    """
    logger = logging.getLogger()
    # Storage for replicated groups
    newgroups = defaultdict(dict)
    for root, cages in groups.items():
        # Position of root (water) (fractional)
        root_pos = waters[root]
        for cage, group_name in cages.items():
            # Position of the cage (fractional)
            cage_pos = cagepos[cage]
            # Relative position of the cage
            delta = rel_wrap(cage_pos - root_pos)
            # (Image) cell that the cage resides
            gcell = np.floor(root_pos + delta)
            for x in range(rep[0]):
                for y in range(rep[1]):
                    for z in range(rep[2]):
                        r = np.array((x, y, z))
                        # label of the root (water) in the replica
                        newroot = root + len(waters) * \
                            (x + rep[0] * (y + rep[1] * z))
                        # replicated cell in which the cage resides.
                        # modulo by positive number is always positive.
                        cr = (r + gcell) % rep
                        newcage = cage + \
                            len(cagepos) * (cr[0] + rep[0]
                                            * (cr[1] + rep[1] * cr[2]))
                        newcage = int(newcage)
                        newgroups[newroot][newcage] = group_name
                        # logger.info(("root",newroot,"newcage", newcage))
    return newgroups


def replicate_labeldict(labels, nmol, rep):
    newlabels = dict()
    for j in labels:
        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))
                    newlabels[newj] = labels[j]
    return newlabels


def replicate_labels(labels, nmol, rep):
    newlabels = set()
    for j in labels:
        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))
                    newlabels.add(newj)
    return newlabels


def replicate_graph(graph, positions, rep):
    repgraph = dg.IceGraph()
    nmol = positions.shape[0]
    for i, j in graph.edges(data=False):
        # positions in the unreplicated cell
        vec = positions[j] - positions[i]
        delta = np.floor(vec + 0.5).astype(int)
        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    xi = (x + delta[0] + rep[0]) % rep[0]
                    yi = (y + delta[1] + rep[1]) % rep[1]
                    zi = (z + delta[2] + rep[2]) % rep[2]
                    newi = i + nmol * (xi + rep[0] * (yi + rep[1] * zi))
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))
                    if graph[i][j]['fixed']:  # fixed pair
                        repgraph.add_edge(newi, newj, fixed=True)
                    else:
                        # shuffle the bond directions
                        if 0 == random.randint(0, 1):
                            repgraph.add_edge(newi, newj, fixed=False)
                        else:
                            repgraph.add_edge(newj, newi, fixed=False)
    # replicate "ignores" == dopants list in the graph
    repgraph.ignores = replicate_labels(graph.ignores, nmol, rep)
    return repgraph


def replicate_positions(positions, rep):
    repx = positions.copy()
    for m in range(1, rep[0]):
        v = np.array([m, 0, 0])
        repx = np.concatenate((repx, positions + v))
    repy = repx.copy()
    for m in range(1, rep[1]):
        v = np.array([0, m, 0])
        repy = np.concatenate((repy, repx + v))
    repz = repy.copy()
    for m in range(1, rep[2]):
        v = np.array([0, 0, m])
        repz = np.concatenate((repz, repy + v))
    return repz / rep


def parse_cages(cages):
    logger = logging.getLogger()
    cagetype = []
    cagepos = []
    if type(cages) is str:
        for line in cages.split("\n"):
            cols = line.split()
            if len(cols) > 0:
                cagetype.append(cols[0])
                cagepos.append([float(x) for x in cols[1:4]])
        cagepos = np.array(cagepos)
    else:
        # Assume it is a list of list
        # flatten
        c = list(flatten(cages))
        while len(c):
            cagetype.append(c.pop(0))
            cagepos.append(c[:3])
            c = c[3:]
        cagepos = np.array(cagepos)
    cagepos[:] -= np.floor(cagepos[:])
    return cagepos, cagetype




def neighbor_cages_of_dopants(dopants, waters, cagepos, cell):
    """
    Just shows the environments of the dopants
    """
    #logger = logging.getLogger()
    dnei = defaultdict(set)
    for site, name in dopants.items():
        org = waters[site]
        for i, pos in enumerate(cagepos):
            #Displacement (relative)
            a = cell.rel2abs(rel_wrap(pos - org))
            sqdistance = np.dot(a, a)
            if sqdistance < 0.57**2:
                dnei[site].add(i)
                # logger.info((i,cagepos[i]))
    return dnei



# They should be separate plugins in the future.
def butyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md"]]])


def pentyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", ["Md", "Me"]]]])


def propyl(cpos, root, cell, molname):
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", "Mc"]])


def ethyl(cpos, root, cell, molname):
    return Alkyl(cpos, root, cell, molname, ["Ma", "Mb"])


def _2_2_dimethylpropyl(cpos, root, cell, molname):
    """
    2,2-dimethylpropyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", "Mc", "Md", "Me"]])


def _2_3_dimethylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me"], "Mf"]])


def _3_methylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me"]]])


def _3_3_dimethylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me", "Mf"]]])


def Alkyl(cpos, root, cell, molname, backbone):
    """
    put a normal-alkyl group rooted at root toward cpos.
    """
    logger = logging.getLogger()
    # logger.info("  Put butyl at {0}".format(molname))
    v1abs = cell.rel2abs(rel_wrap(cpos - root))
    v1 = v1abs / np.linalg.norm(v1abs)

    origin = cell.rel2abs(root)
    CC = 0.154
    rawatoms = alkyl.alkyl(v1, v1abs*1.5/CC, backbone)

    atoms = []
    for i, atom in enumerate(rawatoms):
        atomname, pos = atom
        atompos = cell.abs_wrapf(pos*CC + origin)
        atoms.append([i, molname, atomname, atompos, 0])
    return atoms




class Lattice():
    def __init__(self,
                 lattice_type=None,
                 density=0,
                 rep=(1, 1, 1),
                 depolarize=True,
                 asis=False,
                 cations=dict(),
                 anions=dict(),
                 spot_guests=dict(),
                 spot_groups=dict(),
                 ):
        self.logger      = logging.getLogger()
        self.lattice_type = lattice_type
        self.rep         = rep
        self.depolarize  = depolarize
        self.asis        = asis
        self.cations     = cations
        self.anions      = anions
        self.spot_guests = spot_guests
        self.spot_groups = spot_groups
        lat = safe_import("lattice", lattice_type)
        # Show the document of the module
        try:
            self.doc = lat.__doc__.splitlines()
        except:
            self.doc = []
        self.doc.append("")
        self.doc.append("Command line: {0}".format(" ".join(sys.argv)))
        for line in self.doc:
            self.logger.info("!!! {0}".format(line))
        # ================================================================
        # waters: positions of water molecules
        #
        self.waters = load_numbers(lat.waters)
        self.logger.debug("Waters: {0}".format(len(self.waters)))
        self.waters = self.waters.reshape((self.waters.size // 3, 3))

        # ================================================================
        # cell: cell dimension
        # celltype: symmetry of the cell
        #   see parse_cell for syntax.
        #
        self.cell = Cell(lat.cell, lat.celltype)
        #self.cell = parse_cell(lat.cell, lat.celltype)

        # ================================================================
        # coord: "relative" or "absolute"
        #   Inside genice, molecular positions are always treated as "relative"
        #
        if lat.coord == "absolute":
            self.waters = self.cell.abs2rel(self.waters)
        self.waters = np.array([w - np.floor(w) for w in self.waters])

        # ================================================================
        # pairs: specify the pairs of molecules that are connected.
        #   Bond orientation will be shuffled later
        #   unless it is "fixed".
        #
        self.pairs = None
        try:
            if type(lat.pairs) is str:
                lines = lat.pairs.split("\n")
                self.pairs = []
                for line in lines:
                    columns = line.split()
                    if len(columns) == 2:
                        i, j = [int(x) for x in columns]
                        self.pairs.append((i, j))
            elif type(lat.pairs) is list:
                self.pairs = lat.pairs
                # for pair in lat.pairs:
                #    self.pairs.append(pair)
        except AttributeError:
            self.logger.info("Graph is not defined.")

        # ================================================================
        # bondlen: specify the bond length threshold.
        #   This is used when "pairs" are not specified.
        #   It is applied to the original positions of molecules (before density setting).
        #
        self.bondlen = None
        try:
            self.bondlen = lat.bondlen
            self.logger.info("Bond length (specified): {0}".format(self.bondlen))
        except AttributeError:
            self.bondlen = 1.1 * shortest_distance(self.waters, self.cell)
            self.logger.info("Bond length (assumed): {0}".format(self.bondlen))
        # Set density
        mass = 18  # water
        NB = 6.022e23
        nmol = self.waters.shape[0]  # nmol in a unit cell
        volume = self.cell.volume()  # volume of a unit cell in nm**3
        density0 = mass * nmol / (NB * volume * 1e-21)
        if density <= 0:
            try:
                self.density = lat.density
            except AttributeError:
                self.logger.info(
                    "Density is not specified. Assume the density from lattice.")
                dmin = shortest_distance(self.waters, self.cell)
                self.logger.info(
                    "Closest pair distance: {0} (should be around 0.276 nm)".format(dmin))
                self.density = density0 / (0.276 / dmin)**3
                # self.density = density0
        else:
            self.density = density
        self.logger.info("Target Density: {0}".format(self.density))
        self.logger.info("Original Density: {0}".format(density0))

        # scale the cell according to the (specified) density
        ratio = (density0 / self.density)**(1.0 / 3.0)
        self.cell.scale(ratio)
        if self.bondlen is not None:
            self.bondlen *= ratio
        self.logger.info("Bond length (scaled, nm): {0}".format(self.bondlen))

        # ================================================================
        # double_network: True or False
        #   This is a special option for ices VI and VII that have
        #   interpenetrating double network.
        #   GenIce's fast depolarization algorithm fails in some case.
        #
        try:
            self.double_network = lat.double_network
        except AttributeError:
            self.double_network = False

        # ================================================================
        # cages: positions of the centers of cages
        #   In fractional coordinate.
        #
        self.cagepos = None
        self.cagetype = None
        if "cages" in lat.__dict__:
            self.cagepos, self.cagetype = parse_cages(lat.cages)

        # ================================================================
        # fixed: specify the bonds whose directions are fixed.
        #   you can specify them in pairs at a time.
        #   You can also leave it undefined.
        #
        try:
            if type(lat.fixed) is str:
                lines = lat.fixed.split("\n")
                self.fixed = []
                for line in lines:
                    columns = line.split()
                    if len(columns) == 2:
                        i, j = [int(x) for x in columns]
                        self.fixed.append((i, j))  # Is a tuple
            elif type(lat.fixed) is list:
                self.fixed = []
                for pair in lat.fixed:
                    self.fixed.append(tuple(pair[:2]))  # Must be a tuple
        except AttributeError:
            self.fixed = []
        if "dopeIonsToUnitCell" in lat.__dict__:
            self.dopeIonsToUnitCell = lat.dopeIonsToUnitCell
        else:
            self.dopeIonsToUnitCell = None
        self.dopants = set()
        # if asis, make pairs to be fixed.
        if self.asis and len(self.fixed) == 0:
            self.fixed = self.pairs

        # filled cages
        self.filled_cages = set()
        # groups info
        self.groups = defaultdict(dict)

        # groups for the semi-guest
        # experimental; there are many variation of semi-guest inclusion.
        self.groups_placer = {"Bu-": butyl,
                              "Butyl-": butyl,
                              "Pentyl-": pentyl,
                              "Propyl-": propyl,
                              "2,2-dimethylpropyl-": _2_2_dimethylpropyl,
                              "2,3-dimethylbutyl-": _2_3_dimethylbutyl,
                              "3,3-dimethylbutyl-": _3_3_dimethylbutyl,
                              "3-methylbutyl-": _3_methylbutyl,
                              "Ethyl-": ethyl}


    def format(self, water_type, guests, formatter):
        if 0 in formatter.hooks:
            formatter.hooks[0](self)
        if max(0,*formatter.hooks.keys()) < 1:
            return
        self.stage1()
        if 1 in formatter.hooks:
            formatter.hooks[1](self)
        if max(0,*formatter.hooks.keys()) < 2:
            return
        res = self.stage2()
        if 2 in formatter.hooks:
            formatter.hooks[2](self)
        if max(0,*formatter.hooks.keys()) < 3:
            return
        self.stage3()
        if 3 in formatter.hooks:
            formatter.hooks[3](self)
        if max(0,*formatter.hooks.keys()) < 4:
            return
        self.stage4()
        if 4 in formatter.hooks:
            formatter.hooks[4](self)
        if max(0,*formatter.hooks.keys()) < 5:
            return
        self.stage5()
        if 5 in formatter.hooks:
            formatter.hooks[5](self)
        if max(0,*formatter.hooks.keys()) < 6:
            return
        self.stage6(water_type)
        if 6 in formatter.hooks:
            formatter.hooks[6](self)
        if max(0,*formatter.hooks.keys()) < 7:
            return
        self.stage7(guests)
        if 7 in formatter.hooks:
            formatter.hooks[7](self)

        
    def stage1(self):
        """
        replicate water molecules to make a repeated cell
        """
        self.logger.info("Stage1: Replication.")
        self.reppositions = replicate_positions(self.waters, self.rep)
        # This must be done before the replication of the cell.
        self.logger.info("  Number of water molecules: {0}".format(
            len(self.reppositions)))
        self.graph = self.prepare_random_graph(self.fixed)
        # scale the cell
        self.repcell = Cell(self.cell)
        self.repcell.scale2(self.rep)

        if self.cagepos is not None:
            self.logger.info("  Hints:")
            self.repcagepos = replicate_positions(self.cagepos, self.rep)
            nrepcages = self.repcagepos.shape[0]
            self.repcagetype = [self.cagetype[i % len(self.cagetype)]
                               for i in range(nrepcages)]
            self.cagetypes = defaultdict(set)
            for i, typ in enumerate(self.repcagetype):
                self.cagetypes[typ].add(i)
            # INFO for cage types
            self.logger.info("    Cage types: {0}".format(list(self.cagetypes)))
            for typ, cages in self.cagetypes.items():
                self.logger.info("    Cage type {0}: {1}".format(typ, cages))
            # Up here move to stage 1.
        
        self.logger.info("Stage1: end.")

    def stage2(self):
        """
        make a random graph and replicate.
        """
        self.logger.info("Stage2: Graph preparation.")
        # Some edges are directed when ions are doped.
        if self.dopeIonsToUnitCell is not None:
            self.dopeIonsToUnitCell(self)  # may be defined in the plugin
        # Replicate the dopants in the unit cell
        self.dopants = replicate_labeldict(
            self.dopants, len(self.waters), self.rep)
        self.groups = replicate_groups(
            self.groups, self.waters, self.cagepos, self.rep)
        # self.groups_info(self.groups)
        for root, cages in self.groups.items():
            self.filled_cages |= set(cages)
        # self.logger.info(("filled",self.filled_cages))
        # Replicate the graph
        self.graph = replicate_graph(self.graph, self.waters, self.rep)
        # Dope ions by options.
        if len(self.anions) > 0:
            self.logger.info("  Anionize: {0}.".format(self.anions))
            for site, name in self.anions.items():
                self.graph.anionize(site)
                self.dopants[site] = name
        if len(self.cations) > 0:
            self.logger.info("  Cationize: {0}.".format(self.cations))
            for site, name in self.cations.items():
                self.graph.cationize(site)
                self.dopants[site] = name
        # Count bonds
        nrandom = 0
        nfixed = 0
        for i, j, data in self.graph.edges(data=True):
            if self.graph[i][j]['fixed']:  # fixed pair
                nfixed += 1
            else:
                nrandom += 1
        self.logger.info(
            "  Number of pre-oriented hydrogen bonds: {0}".format(nfixed))
        self.logger.info(
            "  Number of unoriented hydrogen bonds: {0}".format(nrandom))
        self.logger.info("  Number of hydrogen bonds: {0} (regular num: {1})".format(
            nfixed + nrandom, len(self.reppositions) * 2))
        # test2==True means it is a z=4 graph.
        self.test2 = self.test_undirected_graph(self.graph)
        if not self.test2:
            self.logger.info("Test2 failed.")

        self.logger.info("Stage2: end.")
        return self.test2

    def stage3(self):
        """
        make a true ice graph.
        """
        self.logger.info("Stage3: Bernal-Fowler rule.")
        if self.asis:
            self.logger.info("  Skip applying the ice rule by request.")
        else:
            self.graph.purge_ice_defects()
        self.logger.info("Stage3: end.")

    def stage4(self):
        """
        depolarize.
        """
        self.logger.info("Stage4: Depolarization.")
        if not self.depolarize or self.asis:
            self.logger.info("  Skip depolarization by request.")
            self.yapresult = ""
            self.spacegraph = dg.SpaceIceGraph(self.graph,
                                               coord=self.reppositions,
                                               ignores=self.graph.ignores)
        else:
            if self.double_network:
                if (self.rep[0] % 2 == 0) and (self.rep[1] % 2 == 0) and (self.rep[2] % 2 == 0):
                    pass
                else:
                    self.logger.error(
                        "In making the ice structure having the double network (e.g. ices 6 and 7), all the repetition numbers (--rep) must be even.")
                    sys.exit(1)
            self.spacegraph = dg.SpaceIceGraph(self.graph,
                                               coord=self.reppositions,
                                               ignores=self.graph.ignores)
            draw = dg.YaplotDraw(
                self.reppositions, self.repcell.mat, data=self.spacegraph)
            self.yapresult = dg.depolarize(
                self.spacegraph, self.repcell.mat, draw=draw)
        self.logger.info("Stage4: end.")

    def stage5(self):
        """
        orientations for rigid molecules.
        """
        self.logger.info("Stage5: Orientation.")
        # Add small noises to the molecular positions
        if not self.asis:
            self.reppositions += self.repcell.abs2rel(np.random.random(self.reppositions.shape)* 0.01 - 0.005)
        # determine the orientations of the water molecules based on edge
        # directions.
        self.rotmatrices = orientations(
            self.reppositions, self.spacegraph, self.repcell)

        # Activate it.
        # logger.info("The network is not specified.  Water molecules will be orinented randomly.")
        # rotmatrices = [rigid.rand_rotation_matrix() for pos in positions]
        self.logger.info("Stage5: end.")

    def stage6(self, water_type):
        """
        arrange water atoms and replacements
        """
        self.logger.info("Stage6: Atomic positions of water.")
        # assert audit_name(water_type), "Dubious water name: {0}".format(water_type)
        # water = importlib.import_module("genice.molecules."+water_type)
        water = safe_import("molecule", water_type)
        self.atoms = arrange_atoms(self.reppositions,
                                   self.repcell,
                                   self.rotmatrices,
                                   water.sites,
                                   water.labels,
                                   water.name,
                                   ignores=set(self.dopants))
        self.logger.info("Stage6: end.")

    def stage7(self, guests):
        """
        arrange guest atoms
        """
        self.logger.info("Stage7: Atomic positions of the guest.")
        if self.cagepos is not None:
            # the cages around the dopants.
            dopants_neighbors = self.dopants_info(
                self.dopants, self.reppositions, self.repcagepos, self.repcell)
            # put the (one-off) groups
            if len(self.spot_groups) > 0:
                # process the -H option
                for cage, group_to in self.spot_groups.items():
                    group, root = group_to.split(":")
                    self.add_group(cage, group, int(root))
            molecules = defaultdict(list)
            if len(self.spot_guests) > 0:
                # process the -G option
                for cage, molec in self.spot_guests.items():
                    molecules[molec].append(cage)
                    self.filled_cages.add(cage)
            if guests is not None:
                # process the -g option
                for arg in guests:
                    cagetype, spec = arg[0].split("=")
                    assert cagetype in self.cagetypes, "Nonexistent cage type: {0}".format(
                        cagetype)
                    resident = dict()
                    rooms = list(self.cagetypes[cagetype] - self.filled_cages)
                    for room in rooms:
                        resident[room] = None
                    # spec contains a formula consisting of "+" and "*"
                    contents = spec.split("+")
                    vacant = len(rooms)
                    for content in contents:
                        if "*" in content:
                            molec, frac = content.split("*")
                            frac = float(frac)
                        else:
                            molec = content
                            frac = 1.0
                        nmolec = int(frac * len(rooms) + 0.5)
                        vacant -= nmolec
                        assert vacant >= 0, "Too many guests."
                        remain = nmolec
                        movedin = []
                        while remain > 0:
                            r = random.randint(0, len(rooms) - 1)
                            room = rooms[r]
                            if resident[room] is None:
                                resident[room] = molec
                                molecules[molec].append(room)
                                movedin.append(room)
                                remain -= 1
                        #self.logger.info(
                        #    "    {0} * {1} @ {2}".format(molec, nmolec, movedin))
            # Now ge got the address book of the molecules.
            if len(molecules):
                self.logger.info("  Summary of guest placements:")
                self.guests_info(self.cagetypes, molecules)
            if len(self.spot_groups) > 0:
                self.logger.info("  Summary of groups:")
                self.groups_info(self.groups)
            # semi-guests
            for root, cages in self.groups.items():
                assert root in self.dopants
                name = self.dopants[root]
                molname = "G{0}".format(root)
                pos = self.reppositions[root]
                rot = self.rotmatrices[root]
                self.atoms.append([0, molname, name, self.repcell.rel2abs(pos), 0])
                del self.dopants[root]  # processed.
                self.logger.debug((root,cages,name,molname,pos,rot))
                for cage, group in cages.items():
                    assert group in self.groups_placer
                    assert cage in dopants_neighbors[root]
                    cpos = self.repcagepos[cage]
                    self.atoms += self.groups_placer[group](cpos,
                                                            pos,
                                                            self.repcell,
                                                            molname)
            # molecular guests
            for molec, cages in molecules.items():
                gmol = safe_import("molecule", molec)
                cpos = [self.repcagepos[i] for i in cages]
                cmat = [np.identity(3) for i in cages]
                self.atoms += arrange_atoms(cpos, self.repcell,
                                            cmat, gmol.sites, gmol.labels, gmol.name)
        # Assume the dopant is monatomic and replaces one water molecule
        atomset = defaultdict(set)
        for label, name in self.dopants.items():
            atomset[name].add(label)
        for name, labels in atomset.items():
            pos = [self.reppositions[i] for i in sorted(labels)]
            rot = [self.rotmatrices[i] for i in sorted(labels)]
            self.atoms += arrange_atoms(pos,
                                        self.repcell,
                                        rot,
                                        [[0., 0., 0.], ],
                                        [name],
                                        name)
        self.logger.info("Stage7: end.")


    def prepare_random_graph(self, fixed):
        if self.pairs is None:
            self.logger.info("  Pairs are not given explicitly.")
            self.logger.info(
                "  Start estimating the bonds according to the pair distances.")
            # make bonded pairs according to the pair distance.
            # make before replicating them.
            grid = pl.determine_grid(self.cell.mat, self.bondlen)
            assert np.product(grid) > 0, "Too thin unit cell. Consider use of --rep option if the cell was made by cif2ice."
            self.pairs = [v for v in pl.pairlist_fine(
                self.waters, self.bondlen, self.cell.mat, grid, distance=False)]
            # Check using a simpler algorithm.
            if self.logger.level <= logging.DEBUG:
                pairs2 = [v for v in pl.pairlist_crude(
                    self.waters, self.bondlen, self.cell.mat, distance=False)]
                self.logger.debug("pairs: {0}".format(len(self.pairs)))
                self.logger.debug("pairs2: {0}".format(len(pairs2)))
                for pair in self.pairs:
                    i, j = pair
                    assert (i, j) in pairs2 or (j, i) in pairs2
                for pair in pairs2:
                    i, j = pair
                    assert (i, j) in self.pairs or (j, i) in self.pairs

        graph = dg.IceGraph()
        for i, j in fixed:
            graph.add_edge(i, j, fixed=True)
        # Fixed pairs are default.
        for pair in self.pairs:
            i, j = pair
            if graph.has_edge(i, j) or graph.has_edge(j, i):
                pass
            else:
                if random.randint(0, 1) == 0:
                    graph.add_edge(i, j, fixed=False)
                else:
                    graph.add_edge(j, i, fixed=False)
        return graph

    def test_undirected_graph(self, graph):
        # Test
        undir = graph.to_undirected()
        for node in range(undir.number_of_nodes()):
            if node not in undir:
                self.logger.debug("z=0 at {0}".format(node))
            else:
                z = len(list(undir.neighbors(node)))
                if z != 4:
                    self.logger.debug("z={0} at {1}".format(z, node))
        if graph.number_of_edges() != len(self.reppositions) * 2:
            self.logger.info("Inconsistent number of HBs {0} for number of molecules {1}.".format(
                graph.number_of_edges(), len(self.reppositions)))
            return False
        return True

    def dopants_info(self, dopants=None, waters=None, cagepos=None, cell=None):
        if dopants is None:
            dopants = self.dopants
        if waters is None:
            waters = self.waters
        if cagepos is None:
            cagepos = self.cagepos
        if cell is None:
            cell = self.cell
        dopants_neighbors = neighbor_cages_of_dopants(
            dopants, waters, cagepos, cell)
        for dopant, cages in dopants_neighbors.items():
            self.logger.info(
                "    Cages adjacent to dopant {0}: {1}".format(dopant, cages))
        return dopants_neighbors

    def groups_info(self, groups):
        for root, cages in groups.items():
            for cage, group in cages.items():
                self.logger.info(
                    "    Group {0} of dopant {1} in cage {2}".format(group, root, cage))

    def guests_info(self, cagetypes, molecules):
        for cagetype, cageid in cagetypes.items():
            self.logger.info("    Guests in cage type {0}:".format(cagetype))
            for molec, cages in molecules.items():
                cages = set(cages)
                cages &= cageid
                if len(cages):
                    self.logger.info("      {0} * {1} @ {2}".format(molec, len(cages), cages))
        
    def add_group(self, cage, group, root):
        self.groups[root][cage] = group
        self.filled_cages.add(cage)

    def __del__(self):
        self.logger.info("Completed.")



# coding: utf-8
"""
Crude Cif file format
"""

from math import acos,pi
import numpy as np


def nearly_zero(x):
    return np.dot(x,x) < 1e-10


def hook7(lattice):
    lattice.logger.info("Hook7: Output in CIF format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    cellmat = lattice.repcell.mat

    a = cellmat[0,:]
    b = cellmat[1,:]
    c = cellmat[2,:]
    aL= np.linalg.norm(a)
    bL= np.linalg.norm(b)
    cL= np.linalg.norm(c)
    ab = np.dot(a,b)
    bc = np.dot(b,c)
    ca = np.dot(c,a)
    alpha = acos(bc/(bL*cL)) * 180 / pi
    beta  = acos(ca/(cL*aL)) * 180 / pi
    gamma = acos(ab/(aL*bL)) * 180 / pi
    s = ""
    s += "data_genice_{0}\n".format(lattice.lattice_type)
    s += '#' + "\n#".join(lattice.doc) + "\n"
    s += "_cell_length_a                {0}\n".format(aL*10)
    s += "_cell_length_b                {0}\n".format(bL*10)
    s += "_cell_length_c                {0}\n".format(cL*10)
    s += "_cell_angle_alpha             {0}\n".format(alpha)
    s += "_cell_angle_beta              {0}\n".format(beta)
    s += "_cell_angle_gamma             {0}\n".format(gamma)
    s += "\n"
    if nearly_zero(alpha-90) and nearly_zero(beta-90) and nearly_zero(gamma-90):
        s += "_symmetry_cell_setting        'orthorhombic'\n"
        s += "_symmetry_space_group_name_H-M   'P 1 '\n"
    else:
        s += "_symmetry_cell_setting        'triclinic'\n"  #for now it is always triclinic
        s += "_symmetry_space_group_name_H-M   'P 1 '\n"
    s += """
loop_
  _symmetry_equiv_pos_as_xyz
X,Y,Z

loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
"""
    for i in range(len(lattice.atoms)):
        molorder, resname, atomname, position, order = lattice.atoms[i]
        position = lattice.repcell.abs2rel(position)
        label = atomname+"{0}".format(i)
        s += "{4:>6}{0:>6}{1:10.4f}{2:10.4f}{3:10.4f}\n".format(atomname,position[0],position[1],position[2],label)
    s += "\n"
    print(s,end="")
    lattice.logger.info("Hook7: end.")

hooks = {7:hook7}



# coding: utf-8
"""
Hydrogen bond network in @NGPH format.
"""

import numpy as np


def hook4(lattice):
    lattice.logger.info("Hook4: Output the hydrogen bond network.")

    s = ""
    s += "@NGPH\n"
    s += "{0}\n".format(len(lattice.reppositions))
    for i,j,k in lattice.spacegraph.edges(data=True):
        s += "{0} {1}\n".format(i,j)
    s += "-1 -1\n"
    s = "\n".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook4: end.")

hooks = {4:hook4}



# coding: utf-8
"""
Test suite: Kirkwood G factor
"""

import numpy as np

#It should be expressed as a function of distance. 

def hook5(lattice):
    lattice.logger.info("Hook5: Kirkwood G.")
    zvec = np.array([0.,0.,1.])
    dipoles = np.dot(zvec, lattice.rotmatrices)
    n = dipoles.shape[0]
    lattice.logger.info("  Inner products of the dipoles.")
    # inner products of dipole pairs
    ip = np.zeros((n,n,3))
    for i in (0,1,2): # x,y,z
        x = dipoles[:,i]
        xT = x.reshape((n,1))
        ip[:,:,i] = x*xT  #broadcast product
    ip = np.sum(ip, axis=2).reshape(n*n)
    lattice.logger.info("  Pair distances.")
    # pair distances
    coord = lattice.reppositions
    delta = np.zeros((n,n,3))
    for i in (0,1,2):
        x = coord[:,i]
        xT = x.reshape((n,1))
        d = x - xT
        d -= np.floor( d + 0.5 )  #wrap
        delta[:,:,i] = d
    delta = np.dot(delta, lattice.repcell.mat)
    delta = delta*delta
    delta = np.sum(delta, axis=2)
    delta = np.sqrt(delta).reshape(n*n)
    # print(ip)
    # print(delta)
    # print(lattice.repcell.mat)
    lattice.logger.info("  G(r)")
    mx = np.max(delta)
    x = 0.04
    while x < mx:
        filter = ip[delta<x]
        print("{0:.5f} {1:.5f}".format(x, np.sum(filter)/n))
        x += 0.08
    lattice.logger.info("Hook5: end.")


hooks = {5:hook5}



# coding: utf-8
"""
XMol file format (.xyz)
"""

import numpy as np

from genice import rigid

def hook7(lattice):
    lattice.logger.info("Hook7: Output in XYZ format.")
    s = ""
    s += "{0}\n".format(len(lattice.atoms))
    s += "Generated by GenIce. https://github.com/vitroid/GenIce\n"
    for i in range(len(lattice.atoms)):
        molorder, resname, atomname, position, order = lattice.atoms[i]
        s += "{0:5} {1:9.4f} {2:9.4f} {3:9.4f}\n".format(atomname,position[0]*10,position[1]*10,position[2]*10)
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



#!/usr/bin/env python3

#system modules
import os
import sys
import itertools as it
import logging
#external modules
import numpy as np
from requests import get #requests package
import validators        #validators package
from cif2ice import read_cif

def shortest_distance(atoms):
    dmin = 1e99
    for a1,a2 in it.combinations(atoms,2):
        name1,x1,y1,z1 = a1
        name2,x2,y2,z2 = a2
        d = (x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2
        if d < dmin:
            dmin = d
    return dmin**0.5


def is_unique(L, pos):
    for x in L:
        d = x-pos
        d -= np.floor(d + 0.5)
        if np.dot(d,d) < 0.0000001:
            return False
    return True


def genice_lattice(atoms, box, matchfunc=None):
    global cell, celltype, waters, coord, density
    logger = logging.getLogger()
    filtered = []
    if matchfunc is not None:
        for a in atoms:
            if matchfunc(a[0]):
                filtered.append(a)
    else:
        filtered = atoms
    dmin = shortest_distance(filtered)
    scale = 2.76 / dmin

    celltype = "triclinic"
    if (len(box) == 6):
        cell = np.array([[box[0],0,0],[box[1],box[2],0],[box[3],box[4],box[5]]])
    else:
        cell = np.diag(box)
    volume = np.linalg.det(cell)
    icell  = np.linalg.inv(cell)
    uniques = []
    for name,x,y,z in filtered:
        rpos = np.dot([x,y,z], icell)
        rpos -= np.floor(rpos)
        #Do twice to reduce the floating point uncertainty.
        #(Hint: assume the case when x=-1e-33.)
        rpos -= np.floor(rpos)
        if is_unique(uniques, rpos):
            uniques.append(rpos)
    waters = uniques
    coord = "relative"
    # bondlen = 3
    density = len(filtered)*18.0/(volume*scale**3*1e-24*6.022e23)

    
def download(url, file_name):
    # open in binary mode
    with open(file_name, "wb") as file:
        # get request
        response = get(url)
        # write to file
        file.write(response.content)
        
def argparser(arg):
    logger = logging.getLogger()
    args = arg.split(":")
    Nbox = (1,1,1) # should be given as an option
    name = args[0]
    logger.info(__name__)
    #input must be a file......too bad.
    if os.path.exists(name):
        fNameIn = name
    else:
        if validators.url(name):
            URL = name
            name = os.path.basename(name)
            if name[-4:] in (".cif", ".CIF"):
                name = name[:-4]
        elif __name__[-16:] == "lattices.zeolite":
            # it only works when my module name is zeolite.
            URL = "http://www.iza-structure.org/IZA-SC/cif/"+name+".cif"
        fNameIn = name + ".cif"
        assert validators.url(URL)
        download(URL, fNameIn)
    logger.info("Input: {0}".format(fNameIn))
    atoms, box = read_cif.read_and_process(fNameIn, Nbox, make_rect_box=False)
    genice_lattice(atoms, box, matchfunc=lambda x: x[0] != "O")


        
# Do nothing by default; it causes an error when arguments are missing.




def plaintext(a):
    s = ""
    for x in a:
        s += "{0:.4f} ".format(x)
    return s



def Line(v0,v1):
    return "l " + plaintext(v0) + plaintext(v1) + "\n"


def Text(v,txt):
    return "t " + plaintext(v) + " " + txt + "\n"


def Circle(v0):
    return "c " + plaintext(v0) + "\n"


def Arrow(v0,v1):
    return "s " + plaintext(v0) + plaintext(v1) + "\n"



def Polygon(vertices):
    s = "p {0} ".format(len(vertices))
    for v in vertices:
        s += plaintext(v)
    return s + "\n"



def Color(x):
    return "@ {0}\n".format(int(x))


def SetPalette(x,R,G,B):
    return "@ {0} {1} {2} {3}\n".format(int(x),R,G,B)


def Size(x):
    return "r {0}\n".format(float(x))


def Layer(x):
    return "y {0}\n".format(int(x))


def ArrowType(x):
    return "a {0}\n".format(int(x))


def NewPage():
    return "\n"





# coding: utf-8
"""
[1] M. W. Mahoney and W. L. Jorgensen, A five-site model for liquid water and the reproduction of the density anomaly by rigid, nonpolarizable potential functions, J. Chem. Phys. 112 (2000) 8910-8922. [Back]
[2] W. L. Jorgensen, J. Chandrasekhar, J. D. Madura, R. W. Impey, and M. L. Klein, Comparison of simple potential functions for simulating liquid water, J. Chem. Phys. 79 (1983) 926-935
[3] W. L. Jorgensen and J. D. Madura, Temperature and size dependence for monte carlo simulations of TIP4P water, Mol. Phys. 56 (1985) 1381-1392.
"""
import numpy as np
from math import pi,sin,cos

L1 = 0.9572 / 10
L2 = 0.15 / 10
theta=104.52 * pi/180


hy = L1*sin(theta/2)
hz = L1*cos(theta/2)
mz = L2

sites = np.array([[0.0, 0.0, 0.0],
                  [0.0, hy,  hz],
                  [0.0,-hy,  hz],
                  [0.0, 0.0, mz]])
sites -= (sites[1]+sites[2]+sites[3]*16)/18

                  
atoms = ["O","H","H","."]
labels = ["OW","HW1","HW2","MW"]
name = "ICE"





# parallelepiped cell


import numpy as np
import sys
import logging

def rel_wrap(relvec):
    return relvec - np.floor(relvec + 0.5)


def rel_wrapf(relvec):
    return relvec - np.floor(relvec)


class Cell():
    mat = np.zeros(9).reshape(3,3)
    inv = None
    def __init__(self, desc=None, celltype=None):
        if celltype is not None:
            self.parse(desc,celltype)
        else:
            # copy the Cell class
            self.mat = desc.mat.copy()
            self.inv = desc.inv.copy()

            
    def abs2rel(self, absvecs):
        return np.dot(absvecs, self.inv)

        
    def rel2abs(self, relvec):
        return np.dot(relvec, self.mat)

    
    def abs_wrap(self, absvec):
        return self.rel2abs(rel_wrap(self.abs2rel(absvec)))


    def abs_wrapf(self, absvec):
        return self.rel2abs(rel_wrapf(self.abs2rel(absvec)))

    
    def volume(self):
        return np.linalg.det(self.mat)
    

    def scale(self, x):
        self.mat *= x
        self.inv = np.linalg.inv(self.mat)
        

    def scale2(self, x):
        for d in range(3):
            self.mat[d, :] = self.mat[d, :] * x[d]
        self.inv = np.linalg.inv(self.mat)


    def parse(self, desc, celltype):
        logger = logging.getLogger()
        if celltype == "rect":
            if type(desc) is str:
                vec = np.fromstring(desc, sep=" ")
            elif type(desc) is list:
                vec = np.array(desc)
            logger.debug("parse_cell 1: {0}".format(self.mat))
            self.mat = np.diag(vec)
        elif celltype == "monoclinic":
            if type(desc) is str:
                desc = np.fromstring(desc, sep=" ")
            elif type(desc) is list:
                desc = np.array(desc)
            beta = desc[3] * pi / 180.
            M = np.array(((desc[0] * 1.0, desc[1] * 0.0, desc[2] * cos(beta)),
                          (desc[0] * 0.0, desc[1] * 1.0, desc[2] * 0.0),
                          (desc[0] * 0.0, desc[1] * 0.0, desc[2] * sin(beta))))
            # all the vector calculations are done in transposed manner.
            self.mat = M.transpose()
        elif celltype == "triclinic":
            """
            Put the vectors like following:
            cell = "ax 0 0 bx by 0 cx cy cz"
            when you define a unit cell in lattices/
            """
            if type(desc) is str:
                desc = np.fromstring(desc, sep=" ")
                logger.debug(desc)
            elif type(desc) is list:
                desc = np.array(desc)
            self.mat = np.reshape(desc, (3, 3))
            # assert cell[0, 1] == 0 and cell[0, 2] == 0 and cell[1, 2] == 0
        else:
            logger.error("unknown cell type: {0}".format(celltype))
            sys.exit(1)
        self.inv = np.linalg.inv(self.mat)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

####Note: xyz are in coord relative to the cell.


from __future__ import print_function
import math
import itertools as it
import numpy as np
import logging

def Address(pos,grid):
    #residents in each grid cell
    mol = pos % 1 % 1  # avoid cancellation
    return tuple((mol * grid).astype(int))


def ArrangeAddress(xyz,grid):
    #residents in each grid cell
    residents = dict()
    for i in range(len(xyz)):
        address = Address(xyz[i], grid)
        if address not in residents:
            residents[address] = set()
        residents[address].add(i)
    return residents



def pairlist(xyz,grid):
    logger = logging.getLogger()
    logger.debug("START Arrange")
    residents = ArrangeAddress(xyz,grid)
    logger.debug("END Arrange")

    #key-value pairs in the dictionary
    donecellpair = set()
    for address in residents:
        members = residents[address]
        #neighbor cells
        npa = np.array(address)
        k = np.array([(npa + j + grid)%grid for j in range(-1,2)])
        for a2 in it.product(k[:,0],k[:,1],k[:,2]):
            if address == a2:
                if frozenset((address,a2)) not in donecellpair:
                    donecellpair.add(frozenset((address,a2)))
                    for a,b in it.combinations(members,2):
                        yield a,b
            else:
                if a2 in residents:
                    if frozenset((address,a2)) not in donecellpair:
                        donecellpair.add(frozenset((address,a2)))
                        for a in members:
                            for b in residents[a2]:
                                yield a,b


def pairlist_hetero(xyz,xyz2,grid):
    logger = logging.getLogger()
    logger.debug("START Arrange")
    residents  = ArrangeAddress(xyz,grid)
    residents2 = ArrangeAddress(xyz2,grid)
    logger.debug("END Arrange")

    #key-value pairs in the dictionary
    donecellpair = set()
    for address in residents:
        members = residents[address]
        ix,iy,iz = address
        #neighbor cells
        npa = np.array(address)
        k = np.array([(npa + j + grid)%grid for j in range(-1,2)])
        for a2 in it.product(k[:,0],k[:,1],k[:,2]):
            if a2 in residents2:
                if not ((address,a2) in donecellpair):
                    donecellpair.add((address,a2))
                    for a in members:
                        for b in residents2[a2]:
                            yield a,b


                                
#assume xyz and box are numpy.array
def pairlist_fine(xyz,rc,cell,grid,distance=True):
    logger= logging.getLogger()
    for i,j in pairlist(xyz,grid):
        moli = xyz[i]
        molj = xyz[j]
        d = moli-molj
        d -= np.floor( d + 0.5 )
        d = np.dot(d,cell)
        rr = np.dot(d,d)
        if rr < rc**2:
            if distance:
                yield i,j,math.sqrt(rr)
            else:
                yield i,j


def pairlist_crude(xyz,rc,cell,distance=True):
    logger = logging.getLogger()
    logger.debug(xyz)
    logger.debug(rc)
    logger.debug(cell)
    logger.debug(distance)
    for i,j in it.combinations(range(len(xyz)),2):
        moli = xyz[i]
        molj = xyz[j]
        d = moli-molj
        d -= np.floor( d + 0.5 )
        r = np.dot(d,cell)
        rr = np.dot(r,r)
            
        if rr < rc**2:
            logger.debug((d,r,rr,rc**2))
            if distance:
                yield i,j,math.sqrt(rr)
            else:
                yield i,j




#assume xyz and box are numpy.array
def pairlist_fine_hetero(xyz,xyz2,rc,cell,grid,distance=True):
    newpairs = []
    for i,j in pairlist_hetero(xyz,xyz2,grid):
        moli = xyz[i]
        molj = xyz2[j]
        d = moli-molj
        d -= np.floor( d + 0.5 )
        d = np.dot(d,cell)
        rr = np.dot(d,d)
            
        if rr < rc**2:
            if distance:
                newpairs.append((i,j,math.sqrt(rr)))
            else:
                newpairs.append((i,j))
    return newpairs


#
def determine_grid(cell, radius):
    logger = logging.getLogger()
    ct = cell  #.transpose()
    #Cell vectors
    a = ct[0]
    b = ct[1]
    c = ct[2]
    logger.debug("cell a {0}".format(a))
    logger.debug("cell b {0}".format(b))
    logger.debug("cell c {0}".format(c))
    #Edge lengths
    al = np.linalg.norm(a)   #vector length
    bl = np.linalg.norm(b)
    cl = np.linalg.norm(c)
    #Unit vectors of the axes.
    ae = a / al              #unit vectors
    be = b / bl
    ce = c / cl
    #Distance between the parallel faces
    an = np.dot(a,np.cross(be,ce)) #distance to the bc plane
    bn = np.dot(b,np.cross(ce,ae))
    cn = np.dot(c,np.cross(ae,be))
    gf = np.array([an/radius, bn/radius, cn/radius])  # required number of grid cells
    #Check the lengths of four diagonals.
    logger.debug("Grid divisions: {0}".format(np.floor(gf)))
    #print(cell,radius,gf)
    #import sys
    #sys.exit(1)
    return np.floor(gf).astype(int)



def test():
    xyz = []
    for x in range(2):
        for y in range(2):
            for z in range(2):
                xyz.append(np.array((x/100.+1,y/100.+1,z/100.+1)) / 4.)
    xyz2 = []
    for x in range(2):
        for y in range(2):
            for z in range(2):
                xyz2.append(np.array((x/100.+2,y/100.+1,z/100.+1)) / 4.)
    box = np.diag((4,4,4))
    rc = 1.000000001
    grid = determine_grid(box,rc)
    pairs = pairlist_fine_hetero(xyz,xyz2,rc,box,grid)
    for i,j,l in pairs:
        print(i,j,l)
    print(len(pairs))

if __name__ == "__main__":
    test()



# coding: utf-8
"""
Test for generating water positions from cage positions
"""


cages="""
12 0.5 0.5 0.5
14 0.5 0.0 -0.25
14 0.0 0.25 0.5
14 0.75 0.5 0.0
14 0.5 0.0 0.25
12 0.0 0.0 0.0
14 0.25 0.5 0.0
14 0.0 -0.25 0.5
"""



celltype = 'rect'

cell = """
12.747893943706936 12.747893943706936 12.747893943706936
"""

from genice         import FrankKasper
from genice.lattice import parse_cages
from genice.cells   import Cell
cagepos, cagetype = parse_cages(cages)
cell9             = Cell(cell, celltype)
waters = [w for w in FrankKasper.toWater(cagepos, cell9.mat)]
coord = "relative"
density = FrankKasper.estimate_density(waters, cell9.mat, 2.76)
bondlen = 2.76 * 1.2



#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function
from math import *
import sys
import numpy as np
import numpy.linalg
import logging

#Rigid.py uses vertical vectors while other pars of genice uses horizontal vectors.

#v1,v2,v3: numpy arrays
#return value: fourth vector candidate
def decide1(v1,v2,v3):
    v4 = -(v1+v2+v3)
    return v4 / np.linalg.norm(v4)


def decide2(v1,v2):
    d = v1+v2
    d /= np.linalg.norm(d)
    v = np.cross(v1,v2)
    v /= np.linalg.norm(v)
    d *= -sqrt(1./3)
    v *=  sqrt(2./3)
    return d+v,d-v


#outer product of two vectors; return None if vector is too small
def op(i,j,check=True):
    if check and (np.linalg.norm(i) < 0.1 or np.linalg.norm(j) < 0.1):
        return np.array([])
    a = np.cross(i,j)
    if check and np.linalg.norm(a) < 0.1:
        return np.array([])
    return a



#calculate quaternions from a rotation matrix (three orthogonal vectors)
def rotmat2quat0(i,j,k):
    #print sqlen(i),sqlen(j),sqlen(k)
    ex = np.array((1.0, 0.0, 0.0))
    ey = np.array((0.0, 1.0, 0.0))
    ez = np.array((0.0, 0.0, 1.0))
    
    # i軸をx軸に移す回転の軸は、iとxの2分面上にある。
    # j軸をy軸に移す回転の軸は、jとyの2分面上にある。
    # そして、それらを同時にみたす回転の軸は、2つの2分面の交線である。
    # 交線は、2つの面の法線のいずれとも直交する=外積である。*/
    
    a = op(i-ex, j-ey)
    if a.size == 0:
        a = op(i-ex, k-ez)
        if a.size == 0:
            a = op(k-ez, j-ey)
            if a.size == 0:
                #sys.stderr.write("outer prod warning\n")
                #//全く回転しないケース
                return 1.0, 0.0, 0.0, 0.0
    a /= np.linalg.norm(a)
    #/*回転軸aが求まったので、x軸をi軸に重ねる回転の大きさを求める。。*/
    x0 = ex - a[0]*a
    i0 = i  - a[0]*a
    if np.linalg.norm(i0) < 0.1:
        i0 = j - a[1]*a
        x0 = ey - a[1]*a

    i0 /= np.linalg.norm(i0)
    x0 /= np.linalg.norm(x0)
    cosine = np.dot(i0,x0.transpose())
    if cosine < -1.0 or cosine > 1.0:
        cosh = 0.0
        sinh = 1.0
    else:
        cosh=sqrt((1.0+cosine)*0.5)
        sinh=sqrt(1.0-cosh*cosh)
    #//fprintf(stderr,"sinh %24.17e %24.17e %24.17e\n",cosine,cosh,sinh);
    #/*outer product to determine direction*/
    o = op(i0,x0,False)
    a = a.transpose()
    if np.dot(o,a) < 0.0:
        sinh=-sinh;
    
    return np.array((cosh, -sinh*a[0], +sinh*a[1], -sinh*a[2]))



def rotmat2quat(m):
    #print "rotmat2quat is not reliable yet."
    #sys.exit(1)
    n = m.transpose()
    return rotmat2quat0(np.array(n[0]),np.array(n[1]),np.array(n[2]))


def quat2rotmat(q):
    a,b,c,d = q
    sp11=(a*a+b*b-(c*c+d*d))
    sp12=-2.0*(a*d+b*c)
    sp13=2.0*(b*d-a*c)
    sp21=2.0*(a*d-b*c)
    sp22=a*a+c*c-(b*b+d*d)
    sp23=-2.0*(a*b+c*d)
    sp31=2.0*(a*c+b*d)
    sp32=2.0*(a*b-c*d)
    sp33=a*a+d*d-(b*b+c*c)
    return np.array([[sp11,sp12,sp13], [sp21,sp22,sp23], [sp31,sp32,sp33]])



def euler2quat(e):
    ea,eb,ec = e
    a=cos(ea/2)*cos((ec+eb)/2)
    b=sin(ea/2)*cos((ec-eb)/2)
    c=sin(ea/2)*sin((ec-eb)/2)
    d=cos(ea/2)*sin((ec+eb)/2)
    return np.array((a,b,c,d))
    


def quat2euler(q):
    e = np.zeros(3)
    if q[0] == 1.0:
        return e
    p = 2*(q[0]**2+q[3]**2) - 1
    if p>1.0:
        p=1.0
    if p<-1.0:
        p=-1.0
    e[0] = acos(p)
    thh = e[0] / 2.0
    sinthh = sin(thh)
    costhh = cos(thh)
    p = q[0]/costhh
    if p>1.0:
        p=1.0
    if p<-1.0:
        p=-1.0
    p = acos(p)
    if sinthh == 0.0:
        s = 1.0
    else:
        s = q[1]/sinthh
    if s>1.0:
        s=1.0
    if s<-1.0:
        s=-1.0
    s=acos(s)
    if q[3]<0.0:
        p=2*pi-p
    if q[2]>0:
        e[2] = p+s
        e[1] = p-s
    else:
        e[2] = p-s
        e[1] = p+s
    e[1:3] %= (2*pi)
    return e


def qadd(q1,q2):
    a1,b1,c1,d1 = q1
    a2,b2,c2,d2 = q2
    a3=a1*a2-b1*b2-c1*c2-d1*d2
    b3=a1*b2+b1*a2+c1*d2-d1*c2
    c3=a1*c2+c1*a2-b1*d2+d1*b2
    d3=a1*d2+d1*a2+b1*c2-c1*b2
    if a3 < 0:
       a3 = -a3
       b3 = -b3
       c3 = -c3
       d3 = -d3
    return np.array([a3,b3,c3,d3])



def qmul(q,x):
    if q[0] >= 1.0:
        return np.array(q)
    phi = acos(q[0])
    sine = sqrt(1.0-q[0]**2)
    if q[1] < 0:
        phi=-phi
        sine=-sine
    phi *= x
    sine = sin(phi) / sine
    a = np.zeros(4)
    a[0] = cos(phi)
    a[1] = q[1]*sine
    a[2] = q[2]*sine
    a[3] = q[3]*sine
    return a


def test_rotation():
    e = np.array((0.2,0.3,0.4))
    print(e)
    print(quat2euler(euler2quat(e)))
    print(euler2quat(e))
    print(rotmat2quat(quat2rotmat(euler2quat(e))))
    q = euler2quat(e)
    q = qmul(q, 0.5)
    print(qadd(q,q))


    

def test():
    """Testing Docstring"""
    q1 = np.array([0.5,0.5,0.5,0.5]);
    q2 = np.array([0.5,0.5,0.5,-0.5]);
    q21 = qadd(q2,qmul(q1,-1))
    print(q21)
    print(qadd(q21,q1))
    q3 = np.array([0.60876143, -0.45804276,  0.45804276, -0.45804276])
    print(qadd(q3,q1))
    print(qadd(q1,q3))
    test_rotation()
    #first molecule by genice 1h -r 1 1 1 --format q
    q4 = np.array([0.0692,-0.2480,0.2911,-0.9214])
    e4 = quat2euler(q4)
    print(e4)
    qe4 = euler2quat(e4)
    print(qe4)


#http://blog.lostinmyterminal.com/python/2015/05/12/random-rotation-matrix.html
def rand_rotation_matrix(deflection=1.0, randnums=None):
    """
    Creates a random rotation matrix.
    
    deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
    rotation. Small deflection => small perturbation.
    randnums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
    """
    # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c
    
    if randnums is None:
        randnums = np.random.uniform(size=(3,))
        
    theta, phi, z = randnums
    
    theta = theta * 2.0*deflection*np.pi  # Rotation about the pole (Z).
    phi = phi * 2.0*np.pi  # For direction of pole deflection.
    z = z * 2.0*deflection  # For magnitude of pole deflection.
    
    # Compute a vector V used for distributing points over the sphere
    # via the reflection I - V Transpose(V).  This formulation of V
    # will guarantee that if x[1] and x[2] are uniformly distributed,
    # the reflected points will be uniform on the sphere.  Note that V
    # has length sqrt(2) to eliminate the 2 in the Householder matrix.
    
    r = np.sqrt(z)
    Vx, Vy, Vz = V = (
        np.sin(phi) * r,
        np.cos(phi) * r,
        np.sqrt(2.0 - z)
        )
    
    st = np.sin(theta)
    ct = np.cos(theta)
    
    R = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))
    
    # Construct the rotation matrix  ( V Transpose(V) - I ) R.
    
    M = (np.outer(V, V) - np.eye(3)).dot(R)
    return M



if __name__ == "__main__":
    test()



# coding: utf-8
"""
MDView file format
"""

import numpy as np

au = 0.052917721092 # nm
AA = 0.1 # nm

def hook7(lattice):
    lattice.logger.info("Hook7: Output in MDView format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    if __name__[-6:] == 'mdv_au':
        conv = 1.0 / au
    else:
        conv = 10.0
    s = ""
    s += "-center 0 0 0\n"
    s += "-fold\n"
    s += "{0}\n".format(len(lattice.atoms))
    for i in range(len(lattice.atoms)):
        molorder, resname, atomname, position, order = lattice.atoms[i]
        s += "{0:5} {1:9.4f} {2:9.4f} {3:9.4f}\n".format(atomname,*(position[:3]*conv))
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



# coding: utf-8
"""
Direct visualization with VPython.
"""

from collections import defaultdict
import colorsys
import numpy as np
import networkx as nx
import vpython as vp
from countrings import countrings_nx as cr

hue_sat = {3:(60., 1.0), 4:(180, 0.8), 5:(0, 0.5), 6:(0, 0.0), 7:(240, 0.5), 8:(300, 0.5)}

sun = np.array([1., -10., 5.])  # right, down, front
sun /= np.linalg.norm(sun)

def face(center, rpos):
    n = rpos.shape[0]
    
    # normalize relative vectors
    normalized = np.zeros_like(rpos)
    for i in range(n):
        normalized[i] = rpos[i] / np.linalg.norm(rpos[i])
    #normal for each triangle
    normals = np.zeros_like(rpos)
    for i in range(n):
        normals[i] = np.cross(normalized[i-1], normalized[i])
    # central normal
    c_normal = np.sum(normals, axis=0)
    c_normal /= np.linalg.norm(c_normal)
    if np.dot(c_normal, sun) < 0.0:
        c_normal = - c_normal
        normals  = - normals
    
    hue, sat = hue_sat[n]
    bri = 1
    r,g,b = colorsys.hsv_to_rgb(hue/360., sat, bri)
    pos = rpos + center
    v_center = vp.vertex( pos=vp.vector(*center), normal=vp.vector(*c_normal), color=vp.vector(r,g,b))
    vertices = [vp.vertex( pos=vp.vector(*p), normal=vp.vector(*(normals[i])), color=vp.vector(r,g,b) ) for i,p in enumerate(pos)]
    faces = set()
    for i in range(n):
        faces.add(vp.triangle(v0=vertices[i-1], v1=vertices[i], v2=v_center,  ))
#    group.add(sw.shapes.Polygon(pos, fill=rgb, stroke_linejoin="round", fill_opacity="1.0", stroke="#444", stroke_width=3))
    return faces


def hook4(lattice):
    lattice.logger.info("Hook4: VPython (polyhedral expressions).")
    graph = nx.Graph(lattice.spacegraph) #undirected
    cellmat = lattice.repcell.mat
    lattice.vpobjects = defaultdict(set)
    for ring in cr.CountRings(graph).rings_iter(8):
        deltas = np.zeros((len(ring),3))
        d2 = np.zeros(3)
        for k,i in enumerate(ring):
            d = lattice.reppositions[i] - lattice.reppositions[ring[0]]
            d -= np.floor(d+0.5)
            deltas[k] = d
            dd = lattice.reppositions[ring[k]] - lattice.reppositions[ring[k-1]]
            dd -= np.floor(dd+0.5)
            d2 += dd
        # d2 must be zeros
        if np.all(np.absolute(d2) < 1e-5):
            comofs = np.sum(deltas, axis=0) / len(ring)
            deltas -= comofs
            com = lattice.reppositions[ring[0]] + comofs
            com -= np.floor(com)
            # rel to abs
            com    = np.dot(com-0.5,    cellmat)
            deltas = np.dot(deltas, cellmat)
            ringsize = '{0}'.format(len(ring))
            lattice.vpobjects[ringsize] |= face(com, deltas)
    lattice.logger.info("Hook4: end.")








def draw(lattice):
    lattice.logger.info("Hook6: Display water molecules with VPython.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    cellmat = lattice.repcell.mat
    offset = (cellmat[0] + cellmat[1] + cellmat[2]) / 2
    # prepare the reverse dict
    waters = defaultdict(dict)
    for atom in lattice.atoms:
        resno, resname, atomname, position, order = atom
        if "O" in atomname:
            waters[order]["O"] = position - offset
        elif "H" in atomname:
            if "H0" not in waters[order]:
                waters[order]["H0"] = position - offset
            else:
                waters[order]["H1"] = position - offset
    for order, water in waters.items():
        O = water["O"]        
        H0 = water["H0"]        
        H1 = water["H1"]        
        lattice.vpobjects['w'].add(vp.simple_sphere(radius=0.03, pos=vp.vector(*O), color=vp.vector(1,0,0)))
        lattice.vpobjects['w'].add(vp.simple_sphere(radius=0.02, pos=vp.vector(*H0), color=vp.vector(0,1,1)))
        lattice.vpobjects['w'].add(vp.simple_sphere(radius=0.02, pos=vp.vector(*H1), color=vp.vector(0,1,1)))
        lattice.vpobjects['w'].add(vp.cylinder(radius=0.015, pos=vp.vector(*O), axis=vp.vector(*(H0-O))))
        lattice.vpobjects['w'].add(vp.cylinder(radius=0.015, pos=vp.vector(*O), axis=vp.vector(*(H1-O))))
        lattice.vpobjects['l'].add(vp.label(pos=vp.vector(*O), xoffset=30, text="{0}".format(order), visible=False))
    for i,j in lattice.spacegraph.edges(data=False):
        if i in waters and j in waters:  # edge may connect to the dopant
            O = waters[j]["O"]
            H0 = waters[i]["H0"]
            H1 = waters[i]["H1"]
            d0 = H0 - O
            d1 = H1 - O
            rr0 = np.dot(d0,d0)
            rr1 = np.dot(d1,d1)
            if rr0 < rr1 and rr0 < 0.27**2:
                lattice.vpobjects['a'].add(vp.arrow(shaftwidth=0.015, pos=vp.vector(*H0), axis=vp.vector(*(O-H0)), color=vp.vector(1,1,0)))
            if rr1 < rr0 and rr1 < 0.245**2:
                lattice.vpobjects['a'].add(vp.arrow(shaftwidth=0.015, pos=vp.vector(*H1), axis=vp.vector(*(O-H1)), color=vp.vector(1,1,0)))
    lattice.logger.info("  Tips: use keys to draw/hide layers. [3 4 5 6 7 8 a w l]")
    lattice.logger.info("  Tips: Type ctrl-C twice at the terminal to stop.")
    lattice.logger.info("Hook6: end.")


def hook6(lattice):
    draw(lattice)
    # Toggle visibility
    visible = dict()
    for L in lattice.vpobjects:
        visible[L] = False
    visible['l'] = True
    while True:
        ev = vp.scene.waitfor("keydown")
        if ev.key in lattice.vpobjects:
            for L in lattice.vpobjects[ev.key]:
                L.visible = visible[ev.key]
            visible[ev.key] = not visible[ev.key]



hooks = {6:hook6, 4:hook4}



# coding: utf-8
"""
re-make python module for GenIce
"""

import numpy as np
import itertools as it
from math import floor
import re
MAXCELL=11
torr = 1e-8



def isZero(x):
    return -torr < x < torr



def FlagEquivCells(nv, flags, ijk):
    if nv not in flags:
        for d in it.product((-2,-1,0,1,2), repeat=3):
            vv = tuple(np.dot(d,ijk)+nv)
            flags.add(vv)


def FindEmptyCells(cellmat, ijk, relpositions, labels=None):
    newcell = np.dot(ijk, cellmat)
    newcelli = np.linalg.inv(newcell)
    L1 = np.linalg.norm(newcell[0])
    L2 = np.linalg.norm(newcell[1])
    L3 = np.linalg.norm(newcell[2])
    rL = np.array([L1,L2,L3])
    # print(rL)
    ncell = 0
    flags = set()
    queue = [(0,0,0)]
    s = ""
    while len(queue) > 0:
        nv = queue.pop(0)
        if nv not in flags:
            ncell += 1
            # Place atoms
            for i, xyz in enumerate(relpositions):
                # rel to abs
                xxv = np.dot(xyz + nv, cellmat)
                # inner products with axis vector of the newcell
                pv = np.dot(xxv, newcelli)
                # print(pv,rL)
                pv -= np.floor(pv)
                # print(nv)
                label = ""
                if labels is not None:
                    label = labels[i]
                s += "{3} {0:9.4f} {1:9.4f} {2:9.4f}\n".format(pv[0],pv[1],pv[2], label)
            #print("NV:",nv)
            FlagEquivCells(nv, flags, ijk)
            for xi,yi,zi in it.product((-1,0,1), repeat=3):
                nei = nv[0]+xi,nv[1]+yi,nv[2]+zi
                if nei not in flags:
                    #print("nei", nei)
                    queue.append(nei)
    return ncell, s




# Reshaping matrix (Must be integers)
# for now they are hardcoded.  It will be given as options for the plugin in the future.
# ijk = np.array([[1, 1, 1], [1, -1, 0], [1, 1, -2]])
# ijk = np.array([[2, 0, 1], [0, 1, 0], [-1, 0, 2]])
# ijk = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 2]])
# ijk = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])

#This is default.  No reshaping applied.
ijk = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

def hook1(lattice):
    lattice.logger.info("Hook1: Output as a python module.")
    # Original cell matrix.
    cellmat = lattice.repcell.mat
    lattice.logger.info("  Reshaping the unit cell.")
    lattice.logger.info("    i:{0}".format(ijk[0]))
    lattice.logger.info("    j:{0}".format(ijk[1]))
    lattice.logger.info("    k:{0}".format(ijk[2]))
    # reshaped cell might not be rect.
    newcell = np.dot(ijk, cellmat) 
    # replication ratio.
    vol = abs(np.linalg.det(ijk))
    vol = floor(vol*8192+0.5)/8192
    # Unit orthogonal vectors for the newcell.
    e1 = newcell[0].astype(float)
    e1 /= np.linalg.norm(e1)
    e3 = np.cross(newcell[0], newcell[1]).astype(float)
    e3 /= np.linalg.norm(e3)
    e2 = np.cross(e3,e1)
    R = np.array([e1,e2,e3])
    RI = np.linalg.inv(R)
    # Regularization
    # Let x axis of the newcell be along e1
    # and let y axis of the newcell be on the e1-e2 plane.
    regcell = np.dot(newcell, RI)

    # header
    s = ""
    s += '"""\n'
    s += "\n".join(lattice.doc) + "\n"
    s += "Reshaping the unit cell.\n"
    s += "  i:{0}\n".format(ijk[0])
    s += "  j:{0}\n".format(ijk[1])
    s += "  k:{0}\n".format(ijk[2])
    s += '"""\n'
    
    s += "bondlen={0}\n".format(lattice.bondlen*10)
    s += "coord='relative'\n"
    if isZero(regcell[1,0]) and isZero(regcell[2,0]) and isZero(regcell[2,1]):
        s += "celltype='rect'\n"
        s += "cell='{0:.8f} {1:.8f} {2:.8f}'\n".format(regcell[0,0]*10,regcell[1,1]*10,regcell[2,2]*10)
    else:
        s += "celltype='triclinic'\n"
        s += "cell='"
        for d in range(3):
            s += "{0:.8f} {1:.8f} {2:.8f} ".format(regcell[d,0]*10,regcell[d,1]*10,regcell[d,2]*10)
        s += "'\n"
    # s += "cell='{0} {1} {2}'\n".format(ri,rj,rk)
    s += "density={0}\n".format(lattice.density)
    s += 'waters="""'+"\n"
    lattice.logger.info("  Total number of molecules: {0}".format(vol*len(lattice.reppositions)))
    
    ncell, ss = FindEmptyCells(cellmat, ijk, lattice.reppositions)
    assert vol == ncell

    s += ss + '"""' + "\n\n"

    if lattice.cagepos is not None:
        s += 'cages="""'+"\n"
        ncell, ss = FindEmptyCells(cellmat, ijk, lattice.repcagepos, labels=lattice.repcagetype)
        s += ss + '"""'+"\n\n"
    
    print(s,end="")

    lattice.logger.info("Hook1: end.")


def argparser(arg):
    global ijk
    assert re.match("^[-+0-9,]+$", arg) is not None, "Argument must be nine integers separated by commas."
    ijk = np.array([int(x) for x in arg.split(",")]).reshape(3,3)
        

hooks = {1:hook1}



# coding: utf-8
"""
Gro file format
defined in http://manual.gromacs.org/current/online/gro.html
"""


def hook7(lattice):
    lattice.logger.info("Hook7: Output in Gromacs format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    if len(lattice.atoms) > 99999:
        lattice.logger.warn("  Gromacs fixed format cannot deal with atoms more than 99999. Residue number and atom number are faked.")
    cellmat = lattice.repcell.mat
    s = ""
    s += "Generated by GenIce https://github.com/vitroid/GenIce \n"
    s += "{0}\n".format(len(lattice.atoms))
    molorder = 0
    for i in range(len(lattice.atoms)):
        resno, resname, atomname, position, order = lattice.atoms[i]
        if resno == 0:
            molorder += 1
        if len(lattice.atoms) > 99999:
            s += "{0:5d}{1:5s}{2:>5s}{3:5d}{4:8.3f}{5:8.3f}{6:8.3f}\n".format(9999,resname, atomname, 9999,position[0],position[1],position[2])
        else:
            s += "{0:5d}{1:5s}{2:>5s}{3:5d}{4:8.3f}{5:8.3f}{6:8.3f}\n".format(molorder,resname, atomname, i+1,position[0],position[1],position[2])
    if cellmat[1,0] == 0 and cellmat[2,0] == 0 and cellmat[2,1] == 0:
        s += "    {0} {1} {2}\n".format(cellmat[0,0],cellmat[1,1],cellmat[2,2])
    else:
        assert cellmat[0,1] == 0 and cellmat[0,2] == 0 and cellmat[1,2] == 0
        s += "    {0} {1} {2} {3} {4} {5} {6} {7} {8}\n".format(cellmat[0,0],
                                                                cellmat[1,1],
                                                                cellmat[2,2],
                                                                cellmat[0,1],
                                                                cellmat[0,2],
                                                                cellmat[1,0],
                                                                cellmat[1,2],
                                                                cellmat[2,0],
                                                                cellmat[2,1],
                                                                )
    s += '#' + "\n#".join(lattice.doc) + "\n"
    print(s,end="")
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



# coding: utf-8
import numpy as np

def nearly_zero(x):
    return np.dot(x,x) < 1e-10


def run(lattice, water_type="TIP3P", guests=[]):
    """
    Scigress BDL file format (experimental)
    """
    lattice.stage1()   #replicate the unit cell
    lattice.stage2()   #prepare random graph
    lattice.stage3()   #Make an ice graph
    lattice.stage4()   #Depolarize
    lattice.stage5()   #Orientation
    lattice.stage6(water_type)  #Water atoms
    lattice.stage7B(guests)      #Guest atoms
    #lattice.logger.info("Total number of atoms: {0}".format(len(lattice.atoms)))
    lattice.logger.info("Output in Scigress BDL format.")

    a = lattice.cell[0,:]
    b = lattice.cell[1,:]
    c = lattice.cell[2,:]
    aL= np.linalg.norm(a)
    bL= np.linalg.norm(b)
    cL= np.linalg.norm(c)
    ab = np.dot(a,b)
    bc = np.dot(b,c)
    ca = np.dot(c,a)
    alpha = acos(bc/(bL*cL)) * 180 / pi
    beta  = acos(ca/(cL*aL)) * 180 / pi
    gamma = acos(ab/(aL*bL)) * 180 / pi
    automatic_mass = True
    s = []
    # Lines 1..2: Cell info
    s.append("{0:8.4f}  {1:8.4f}  {2:8.4f}".format(aL*10,bL*10,cL*10))
    s.append("{0:8.4f}  {1:8.4f}  {2:8.4f}".format(alpha,beta,gamma))
    #parse -g options again to determine the number of guest types
    guesttypes = set()
    for arg in guests:
        key, value = arg[0].split("=")
        guesttypes.add(value)
    # Line 3: number of molecular types
    s.append("{0:2}".format(len(guesttypes)+1))
    # Line 4: Guests.
    for gname in lattice.guestAtoms:
        guest = safe_import("molecule", gname)
        #molecule name, number of molecules, sites in a molecule, bonds in a molecule
        #last one is undefined i GenIce.
        s.append("{0:15}                {1:4}  {2:3}  {3:3}".format(gname, lattice.nGuestAtoms[gname], len(guest.sites), 0))
        for i in range(len(lattice.guestAtoms[gname])):
            molorder, resname, atomname, position = lattice.guestAtoms[gname][i]
            q = 0.0  #charge
            s.append("{0:4}  {1:8.4f}  {2:15.9f}  {3:15.9f}  {4:15.9f}".format(atomname,q,position[0],position[1],position[2]))

    #water
    water = safe_import("molecule", water_type)
    s.append("{0:15}                {1:4}  {2:3}  {3:3}".format(water_type, len(lattice.reppositions), len(water.sites), 0))
    for i in range(len(lattice.atoms)):
        molorder, resname, atomname, position = lattice.atoms[i]
        q = 0.0
        s.append("{0:4}  {1:8.4f}  {2:15.9f}  {3:15.9f}  {4:15.9f}".format(atomname,q,position[0],position[1],position[2]))

    for i in range(len(s)):
        line = s[i]
        print("{0:06}    ".format(i+1) + line)

#
#limitations:            
#1)It is a structure file and force field inf should not be included.
#however, BDL requests the partial carge info in it.
#2)Atomic naming rule is always a problem.
#



#!/usr/bin/env python
#coding: utf-8
#input: coordinate of the nodes, the digraph obeying the ice rule.
#output: the digraph with zero net dipole.

from __future__ import print_function
import sys
import math
import networkx
import random
import numpy as np
import logging
from genice import yaplotlib as yp
from genice import pairlist as pl

#Dijkstra


import heapq

def flatten(L):       # Flatten linked list of form [0,[1,[2,[]]]]
    while len(L) > 0:
        yield L[0]
        L = L[1]

#modified from http://code.activestate.com/recipes/119466/
#for networkx
#Extended for multiple goals
def shortest_path(G, start, ends):
    q = [(0, start, ())]  # Heap of (cost, path_head, path_rest).
    visited = set()       # Visited vertices.
    while True:
        try:
            ppap = heapq.heappop(q)
        except IndexError:
            return None
        (cost, v1, path) = ppap
        if v1 not in visited:
            visited.add(v1)
            if v1 in ends:
                return list(flatten(path))[::-1] + [v1]
            path = (v1, path)
            for v2 in G[v1]:
                if v2 not in visited and not G[v1][v2]['fixed']:
                    heapq.heappush(q, (cost + 1, v2, path))





class YaplotDraw(networkx.DiGraph):
    def __init__(self, coord, cell, data=None):
        super().__init__(data)
        self.coord = coord # fractional coord.
        self.cell  = cell

    def draw_edge(self,i,j):
        ci = self.coord[i]  #0..1
        cj = self.coord[j]
        d = cj - ci
        d -= np.floor( d + 0.5 )
        xi = np.dot(ci,self.cell)
        xj = np.dot(ci+d,self.cell)
        if self.has_edge(i,j):
            return yp.Color(4) + yp.ArrowType(2) + yp.Arrow(xi, xj)
        elif self.has_edge(j,i):
            return yp.Color(5) + yp.ArrowType(2) + yp.Arrow(xj, xi)
        else:
            return yp.Color(0) + yp.Line(xi, xj)
            

    def draw_cell(self):
        s = yp.Color(2)
        ex = np.array([1.,0.,0.])
        ey = np.array([0.,1.,0.])
        ez = np.array([0.,0.,1.])
        x = np.dot(ex,self.cell)
        y = np.dot(ey,self.cell)
        z = np.dot(ez,self.cell)
        zero = np.zeros_like(x)
        for vx in (zero, x):
            for vy in (zero, y):
                s += yp.Line(vx+vy,vx+vy+z)
        for vx in (zero, x):
            for vz in (zero, z):
                s += yp.Line(vx+vz,vx+y+vz)
        for vz in (zero, z):
            for vy in (zero, y):
                s += yp.Line(vy+vz,x+vy+vz)
        return s

    def draw_path(self,path):
        s = yp.Color(3)
        for i in range(len(path)-1):
            j,k = path[i],path[i+1]
            s += self.draw_edge(j,k)
        return s
    

    def draw_all(self):
        s = yp.Color(3)
        s += self.draw_cell()
        for i,j in self.edges():
            s += self.draw_edge(i,j)
        return s


class IceGraph(networkx.DiGraph):
    def __init__(self, data=None):
        super(IceGraph, self).__init__(data)
        #set of nodes that ignore the ice rule.
        #is added automatically in cationize/anionize
        self.ignores = set()  

#    def register_pairs(self,pairs):
#        self.clear()
#        for pair in pairs:
#            x,y = pair[:2]
#            self.add_edge(x,y)

    def cationize(self, which):
        invert = set()
        fix    = set()
        for i,j,data in self.edges(data=True):
            if j==which:
                invert.add((i,j))
                fix.add((j,i))
            elif i==which:
                fix.add((i,j))
        for i,j in invert:
            self.invert_edge(i,j)
        for i,j in fix:
            self[i][j]['fixed'] = True
        self.ignores.add(which)

    def anionize(self, which):
        invert = set()
        fix    = set()
        for i,j,data in self.edges(data=True):
            if i==which:
                invert.add((i,j))
                fix.add((j,i))
            elif j==which:
                fix.add((i,j))
        for i,j in invert:
            self.invert_edge(i,j)
        for i,j in fix:
            self[i][j]['fixed'] = True
        self.ignores.add(which)
        
    
    def invert_edge(self,from_,to_):
        """
        also invert the attribute vector
        """
        if not self.has_edge(from_, to_):
            logging.getLogger().error("No edge ({0},{1}).".format(from_,to_))
        assert not self[from_][to_]['fixed']
        self.remove_edge(from_,to_)
        self.add_edge(to_,from_,fixed=False)

        
    def invert_path(self, path):
        for i in range(len(path)-1):
            f = path[i]
            t = path[i+1]
            self.invert_edge(f,t) #also invert the attribute vector
            
    def _goahead(self,node,marks,order):
        while node not in marks:
            marks[node] = len(order)
            order.append(node)
            nei = self.neighbors(node)
            next = random.randint(0,1)
            assert len(nei) == 2, "Dangling bond: {0} {1}".format(node,nei)
            node = nei[next]
        #a cyclic path is found.
        #trim the uncyclic part
        order = order[marks[node]:]
        #add the node at the last again
        order.append(node)
        return order


    def homodromiccycle(self):
        """
        Randomly select a homodromic cycle
        """
        marks = dict()
        order = []
        node = random.randint(0,self.number_of_nodes()-1)
        return self._goahead(node,marks,order)


    def isZ4(self):
        """
        Reply whether all the vertices have four neighbors or not.
        """
        undir = self.to_undirected()
        for node in range(undir.number_of_nodes()):
            if len(list(undir.neighbors(node))) != 4:
                return False
        return True


    def purgedefects(self, defects):
        d = defects[0]
        logger = logging.getLogger()
        logger.debug(self.ignores)
        if d in self.ignores:
            defects.pop(0)
            return
        if self.degree(d) != 4:  # TSL
            assert self.degree(d) < 4
            # logger.warn("  Defect {0} {1} >>{2} <<{3}".format(d,self.degree(d),self.in_degree(d),self.out_degree(d)))
            if self.in_degree(d) <= 2 and self.out_degree(d) <= 2:  #acceptable
                defects.pop(0)
                return
        if self.in_degree(d) == 2 and self.out_degree(d) == 2:
            defects.pop(0)
            return
        if self.in_degree(d) > 2:
            nodes = list(self.predecessors(d))
            i = random.randint(0,len(nodes)-1)
            node = nodes[i]
            if not self[node][d]['fixed']:
                self.invert_edge(node,d)
                defects.append(node)
        if self.out_degree(d) > 2:
            nodes = list(self.successors(d))
            i = random.randint(0,len(nodes)-1)
            node = nodes[i]
            if not self[d][node]['fixed']:
                self.invert_edge(d,node)
                defects.append(node)

            
    def bernal_fowler_defects(self):
        """
        Reply the list of defective vertices.

        It also counts the "ignore_ice_rules" sites.
        """
        logger = logging.getLogger()
        defects = []
        for i in range(self.number_of_nodes()):
            if self.in_degree(i) != 2 or self.out_degree(i) != 2:
                defects.append(i)
        return defects


    def excess_in_defects(self):
        """
        Reply the list of defective vertices.
        """
        for i in range(self.number_of_nodes()):
            if self.in_degree(i) > 2:
                yield i

    def excess_out_defects(self):
        """
        Reply the list of defective vertices.
        """
        for i in range(self.number_of_nodes()):
            if self.out_degree(i) > 2:
                yield i

    
    def purge_ice_defects(self):
        logger = logging.getLogger()
        # TSL
        # if not self.isZ4():
        #    logger.error("Some water molecules do not have four HBs.")
        #    sys.exit(1)
        defects = self.bernal_fowler_defects()
        target = 1
        while len(defects) > target:
            target *= 2
        while len(defects)>0:
            self.purgedefects(defects)
            if len(defects) <= target:
                logger.info("  Defects remaining: {0}".format(len(defects)))
                target //= 2
        # TSL
        # assert set(defects) == self.ignores, "Some water molecules do not obey the ice rule. {0} {1}".format(defects, self.ignores)

                
    def is_homodromic(self, path):
        for i in range(len(path)-1):
            if not self.has_edge(path[i], path[i+1]):
                return False
        return True
            

            


class SpaceIceGraph(IceGraph):
    """
    Digraph with geometrical info
    """
    XAXIS=1
    YAXIS=2
    ZAXIS=3
    def __init__(self, data=None, coord=None, pbc=True, ignores=set()):
        super(SpaceIceGraph, self).__init__(data)
        if coord is not None:
            self.add_vectors(coord, pbc) # fractional coord
        self.ignores = ignores
            
    def add_vectors(self, coord, pbc=True):
        """
        add vector attributes to each edge
        """
        self.coord = coord #Shall it be copied?
        for i,j,k in self.edges(data=True):
            vec = coord[j] - coord[i]
            if pbc:
                vec -= np.floor(vec + 0.5)
            k["vector"] = vec  #each edge has "vector" attribute
        
    def dipole_moment(self, order):
        """
        normally zero for a cycle.
        Non-zero when the cycle goes across the cell.

        For a cycle, the first element of the order must be the same as the last one.
        """
        delta = np.zeros(3)
        for i in range(len(order)-1):
            delta += self.get_edge_data(order[i],order[i+1])["vector"]
        return delta
            
    def invert_edge(self,from_,to_):
        """
        also invert the attribute vector
        """
        if not self.has_edge(from_, to_):
            logging.getLogger().error("No edge ({0},{1}).".format(from_,to_))
        v = self.get_edge_data(from_,to_)["vector"]
        fixed = self.get_edge_data(from_,to_)["fixed"]
        assert not fixed
        self.remove_edge(from_,to_)
        self.add_edge(to_,from_,vector=-v, fixed=False)


        
    def net_polarization(self):
        dipole = np.zeros(3)
        for i,j,k in self.edges(data=True):
            dipole += k["vector"]
        return dipole

    def vector_check(self):
        logger = logging.getLogger()
        for i,j,k in self.edges(data=True):
            if k is None:
                logger.error("The edge ({0},{1}) has no vector.".format(i,j))
                


def find_apsis(coord, cell, distance, vertex, axis):
    logger = logging.getLogger()
    grid = pl.determine_grid(cell, distance)
    logger.debug("Grid: {0}".format(grid))
    #for Z case
    apsis = coord[vertex] + axis*0.5
    #find the atoms near the apsis
    pairs = pl.pairlist_fine_hetero([apsis,], coord, distance, cell, grid, distance=True)
    logger.debug("Neighbors of the apsis: {0}".format(pairs))
    min_d = 1e99
    min_a = -1
    for i,j,d in pairs:
        if d < min_d:
            min_a = j
            min_d = d
    return min_a
        

def estimate_edge_length(spaceicegraph, cell, vertex):
    logger = logging.getLogger()
    # In case an anion is selected by bad fortune.
    if len(spaceicegraph.adj[vertex]) == 0:
        return 0
    nei = list(spaceicegraph.adj[vertex])[0]
    delta = spaceicegraph.coord[vertex] - spaceicegraph.coord[nei]
    delta -= np.floor( delta + 0.5 )
    delta = np.dot(delta, cell)  #distance in abs coord
    distance = np.dot(delta,delta)**0.5
    logger.debug("Distance={0}".format(distance))
    return distance


def traversing_cycle(spaceicegraph, cell, axis, draw=None):
    """
    Find a farthest atom from the given atom, and
    make the shortest paths between them.
    """
    logger = logging.getLogger()

    distance = 0
    while distance == 0:
        vertex = random.randint(0,spaceicegraph.number_of_nodes()-1)
        distance = estimate_edge_length(spaceicegraph, cell, vertex)
    while True:
        vertex = random.randint(0,spaceicegraph.number_of_nodes()-1)
        apsis = find_apsis(spaceicegraph.coord, cell, distance*1.3, vertex, axis)
        logger.debug("Apsis of {0}: {1}".format(vertex, apsis))
        path1 = shortest_path(spaceicegraph, vertex, [apsis,])
        logger.debug("Path1: {0}".format(path1))
        if path1 is None:
            #No path found, probably because of the double networks
            continue
        logger.debug("Dipole of the path1: {0}".format(spaceicegraph.dipole_moment(path1)))
        path2 = shortest_path(spaceicegraph, apsis, [vertex,])
        logger.debug("Path2: {0}".format(path2))
        if path2 is None:
            #No path found, probably because of the double networks
            continue
        logger.debug("Dipole of the path2: {0}".format(spaceicegraph.dipole_moment(path2)))
        #they should make a cycle.
        #It should go across the cell with a probability of 0.5.
        #It should g across the cell in an expected direction
        #with a probability of 0.25.
        #So we need some loops to get the expected one.
        cycle = path1 + path2[1:]
        d = spaceicegraph.dipole_moment(cycle) - axis
        logger.debug("Axis: {0} {1}".format(axis,spaceicegraph.dipole_moment(cycle)))
        rr = np.dot(d,d)
        if rr < 0.1:
            break
    logger.debug("Dipole of the harvest: {0}".format(spaceicegraph.dipole_moment(cycle)))
    return cycle

def depolarize(spaceicegraph, cell, draw=None):
    """
    Find a farthest atom (apsis) from the given atom, and
    make the shortest paths between them.

    It works much better than depolarize()
    """
    logger = logging.getLogger()
    logger.debug("  isZ4: {0}".format(spaceicegraph.isZ4()))
    logger.debug("  defects: {0}".format(spaceicegraph.bernal_fowler_defects()))
    spaceicegraph.vector_check()
    s = "" # for yaplot

    # TSL
    # defect-defect chains
    defects = []
    for node in spaceicegraph.nodes():
        if spaceicegraph.degree(node) != 4:
            assert spaceicegraph.degree(node) < 4
            defects.append(node)
    logger.info("  Non Z4: {0}".format(defects))
    if len(defects) > 0:
        reject_count = 10
        while reject_count > 0:
            net_polar = spaceicegraph.net_polarization()
            # logger.info("  Net polarization: {0}".format(net_polar))
            if np.dot(net_polar, net_polar) < 0.05**2:
                break  # without problem
            while True:
                orig = defects[random.randint(0,len(defects)-1)]
                if spaceicegraph.out_degree(orig) == 2:
                    break
            while True:
                dest = defects[random.randint(0,len(defects)-1)]
                if spaceicegraph.in_degree(dest) == 2:
                    break
            path = shortest_path(spaceicegraph, orig, [dest,])
            if path is None:
                continue
            logger.debug("  Dipole moment = {0}".format(spaceicegraph.dipole_moment(path)))
            new_net_polar = net_polar - spaceicegraph.dipole_moment(path)*2
            if np.linalg.norm(new_net_polar) < np.linalg.norm(net_polar):
                if draw is not None:
                    s += yp.Size(0.03)
                    s += draw.draw_cell()
                    s += draw.draw_path(path)
                    s += yp.NewPage()
                spaceicegraph.invert_path(path)
                ## chk_net_polar = spaceicegraph.net_polarization()
                net_polar = new_net_polar
                logger.info("  Net polarization: {0}".format(net_polar))
                ## assert np.linalg.norm(new_net_polar - chk_net_polar) < 1e-5
                ## logger.debug("New dipole {0}".format(new_net_polar))
                ## logger.debug("Chk dipole {0}".format(chk_net_polar))
                reject_count = 10
            else:
                logger.debug("  Reject inversion")
                reject_count -= 1




    while True:
        net_polar = spaceicegraph.net_polarization()
        logger.info("  Net polarization: {0}".format(net_polar))
        if np.dot(net_polar, net_polar) < 0.2**2:
            break  # without problem
        if -1 <= net_polar[0]  <= 1 and -1 <= net_polar[1]  <= 1 and -1 <= net_polar[2]  <= 1:
            logger.info("  Gave up eliminating the polarization. (2)")
            break
        if net_polar[0] > 1.0:
            logger.debug("Depolarize +X")
            axis = np.array([+1.0, 0.0, 0.0])
        elif net_polar[0] < -1.0:
            logger.debug("Depolarize -X")
            axis = np.array([-1.0, 0.0, 0.0])
        elif net_polar[1] > 1.0:
            logger.debug("Depolarize +Y")
            axis = np.array([0.0,+1.0, 0.0])
        elif net_polar[1] < -1.0:
            logger.debug("Depolarize -Y")
            axis = np.array([0.0,-1.0, 0.0])
        elif net_polar[2] > 1.0:
            logger.debug("Depolarize +Z")
            axis = np.array([0.0, 0.0,+1.0])
        elif net_polar[2] < -1.0:
            logger.debug("Depolarize -Z")
            axis = np.array([0.0, 0.0,-1.0])
        cycle = traversing_cycle(spaceicegraph, cell, axis, draw)
        if cycle is not None:
            edges = [(cycle[i], cycle[i+1]) for i in range(len(cycle)-1)]
            if len(edges) != len(set(edges)):
                logger.debug("The cycle is entangled.")
            else:
                if draw is not None:
                    s += yp.Size(0.03)
                    s += draw.draw_cell()
                    s += draw.draw_path(cycle)
                    s += yp.NewPage()
                spaceicegraph.invert_path(cycle)
                spaceicegraph.vector_check()
    
    
    logger.debug("isZ4: {0}".format(spaceicegraph.isZ4()))
    logger.debug("defects: {0}".format(spaceicegraph.bernal_fowler_defects()))
    return s


def purge_ice_defects(icegraph):
    """
    This is faster than the method in icegraph, but
    it also polarizes the graph in the course of purging.
    """
    logger = logging.getLogger()
    while len(icegraph.bernal_fowler_defects()) > 0:
        logger.info("# of defects: {0}".format(len(icegraph.bernal_fowler_defects())))
        ins = set(icegraph.excess_in_defects())
        for out in icegraph.excess_out_defects():
            while icegraph.out_degree(out) > 2:
                path = shortest_path(icegraph, out, [ins,])
                if path is not None:
                    logger.debug("# of in defects: {0}".format(len(ins)))
                    icegraph.invert_path(path)
                    end = path[-1]
                    if icegraph.in_degree(end) == 2:
                        ins.remove(end)
                #logger.debug("IN:{0}".format(len(set(icegraph.excess_in_defects()))))
                #logger.debug("OUT:{0}".format(len(set(icegraph.excess_out_defects()))))



#!/usr/bin/env python3

#from Frank-Kasper (Tetrahedrally close-packed) alloy to its dual.

#Standard libs
import itertools as it
import logging
from collections import defaultdict

#non-standard libs
import numpy as np

#genice libs
from genice import pairlist as pl
from genice.lattice import flatten

def shortest_distance(atoms, cell):
    logger = logging.getLogger()
    dmin = 1e99
    for a1,a2 in it.combinations(atoms,2):
        d = a1-a2
        d -= np.floor(d + 0.5)
        dv = np.dot(d, cell)
        dd = np.dot(dv,dv)
        if dd < dmin:
            dmin = dd
    logger.debug("shortest_distance: {0}".format(dmin**0.5))
    return dmin**0.5


def estimate_density(atoms, cell, bondlen):
    dmin = shortest_distance(atoms, cell)
    scale = bondlen / dmin
    return 18/6.022e23*len(atoms) / (np.linalg.det(cell)*1e-24 * scale**3)

    


def is_zero(v):
    return np.dot(v,v) < 1e-10


def equivalents(v, cell, rc):
    """
    yield a set of vectors pointing to the image of the original point v.
    """
    origin = v.copy()
    img = [[0., 1.],[0., 1.],[0., 1.]]
    for d in range(3):
        if origin[d] > 0.0:
            origin[d] -= 1.0
    for x in img[0]:
        for y in img[1]:
            for z in img[2]:
                d = origin + np.array([x,y,z])
                r = np.dot(d,cell)
                if np.dot(r,r) < rc**2:
                    yield d


def adjacency_vectors(pairs, rc, coord, cell):
    logger = logging.getLogger()
    vertices = list(set([v for v in flatten(pairs)]))
    adjv = dict()
    adjd = dict()
    for v in vertices:
        adjv[v] = []
        adjd[v] = []
    for i,j in pairs:
        d = coord[j] - coord[i]
        d -= np.floor(d + 0.5)
        for dd in equivalents(d, cell, rc):
            adjd[i].append(dd)
            adjv[i].append(j)
            adjd[j].append(-dd)
            adjv[j].append(i)
    return vertices, adjv, adjd


def tetrahedra(pairs, rc, coord, cell):
    logger = logging.getLogger()
    vertices, adjv, adjd = adjacency_vectors(pairs, rc, coord, cell)
    for v in vertices:
        logger.debug(len(adjv[v]))
        for i,j,k in it.combinations(range(len(adjv[v])), 3):
            vi, vj, vk = adjv[v][i], adjv[v][j], adjv[v][k]
            if vi < v or vj < v or vk < v:
                continue
            di, dj, dk = adjd[v][i], adjd[v][j], adjd[v][k]
            dij = np.dot(di - dj, cell)
            djk = np.dot(dj - dk, cell)
            dki = np.dot(dk - di, cell)
            if np.dot(dij,dij) < rc**2:
                if np.dot(djk,djk) < rc**2:
                    if np.dot(dki,dki) < rc**2:
                        logger.debug((v,vi,vj,vk))
                        yield (v,vi,vj,vk), (coord[v], di,dj,dk)


def toWater(coord, cell):
    logger = logging.getLogger()
    dmin  = shortest_distance(coord,cell)
    pairs = [v for v in pl.pairlist_crude(coord, dmin*1.4, cell, distance=False)]
    for vtet, dtet in tetrahedra(pairs, dmin*1.4, coord, cell):
        p = dtet[0] + (dtet[1] + dtet[2] + dtet[3])/4
        p -= np.floor( p )
        yield p

    
    



# coding: utf-8
"""
towhee file format (.coods?)
"""

import numpy as np

from genice import rigid

def hook7(lattice):
    lattice.logger.info("Hook7: Output in TowHee format.")
    cellmat = lattice.repcell.mat
    s = ""
    for i,atom in enumerate(lattice.atoms):
        molorder, resname, atomname, position, order = atom
        s += "{0:20.15f}{1:20.15f}{2:20.15f}   {3:6}{4:3}{5:6}{6:6}\n".format(position[0]*10,position[1]*10,position[2]*10,atomname,resname,order+1,i+1)
    print(s,end="")
    s = "  Cell Dimensions:"
    lattice.logger.info(s)
    s = "    a: {0} {1} {2}".format(cellmat[0,0]*10,
                                       cellmat[0,1]*10,
                                       cellmat[0,2]*10)
    lattice.logger.info(s)    
    s = "    b: {0} {1} {2}".format(cellmat[1,0]*10,
                                       cellmat[1,1]*10,
                                       cellmat[1,2]*10)
    lattice.logger.info(s)
    s = "    c: {0} {1} {2}".format(cellmat[2,0]*10,
                                       cellmat[2,1]*10,
                                       cellmat[2,2]*10)
    lattice.logger.info(s)
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



# coding: utf-8
"""
Centers of mass of water molecule 
"""

import numpy as np


def hook1(lattice):
    lattice.logger.info("Hook1: Output centers of mass of water molecules.")
    cellmat = lattice.repcell.mat
    s = ""
    if cellmat[1,0] == 0 and cellmat[2,0] == 0 and cellmat[2,1] == 0:
        s += "@BOX3\n"
        s += "{0} {1} {2}\n".format(cellmat[0,0]*10,cellmat[1,1]*10,cellmat[2,2]*10)
    else:
        s += "@BOX9\n"
        for d in range(3):
            s += "{0} {1} {2}\n".format(cellmat[0,d]*10,cellmat[1,d]*10,cellmat[2,d]*10)
    s += "@AR3R\n"
    s += "{0}\n".format(len(lattice.reppositions))
    for rpos in lattice.reppositions:
        s += "{0:9.4f} {1:9.4f} {2:9.4f}\n".format(rpos[0],
                                                rpos[1],
                                                rpos[2])
    s = "\n".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook1: end.")


hooks = {1:hook1}



# coding: utf-8
#Minimal implementation for genice.
#openjscad-like postmodifier style
"""
cell is in nm

openscad2 comes up with OO style
"""

import numpy as np


bondfunc="""
module bond(p1=[0,0,0],p2=[1,1,1],r=1)
{
    d = p2 - p1;
    H = sqrt(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]);
    R = sqrt(d[0]*d[0] + d[1]*d[1]);
    //atan2(y,x)
    theta = 90-atan2(d[2],R);
    phi   = atan2(d[1],d[0]);
    echo(theta,phi);
    translate(p1)
    rotate([0,theta,phi])
    cylinder(r=r, h=H);
}
"""

class OpenScad():
    def __init__(self, s=""):
        self.string = s

    def encode(self, *codes): #Stored value as a string
        return "".join([code.__str__() for code in codes])
    
    def use(self, f):
        return OpenScad("use <{0}>;\n".format(f))

    def defvar(self, name, value):
        return OpenScad("{0}={1};\n".format(name,value))

    def translate(self, value):
        return OpenScad("translate([{0},{1},{2}]){{\n{3}}} //translate\n".format(*value, self.string))

    def scale(self, value):
        return OpenScad("scale([{0},{1},{2}]){{\n{3}}} //scale\n".format(*value, self.string))

    def rotate(self, value):
        return OpenScad("rotate([{0},{1},{2}]){{\n{3}}} //rotate\n".format(*value, self.string))

    def mirror(self, value):
        return OpenScad("mirror([{0},{1},{2}]){{\n{3}}} //mirror\n".format(*value, self.string))

    def add(self, *values):
        return OpenScad("union(){\n" + "".join([self.__str__()] + [value.__str__() for value in values]) + "} //union\n")

    def union(self, *values):
        return self.add(*values)
    
    def subtract(self, *values):
        return OpenScad("difference(){\n" + "".join([self.__str__()] + [value.__str__() for value in values]) + "} //difference\n")

    def intersect(self, *values):
        return OpenScad("intersection(){\n" + "".join([self.__str__()] + [value.__str__() for value in values]) + "} //intersection\n")

    #aliases for backward compat
    def difference(self, *values):
        return self.subtract(*values)
    
    def intersection(self, *values):
        return self.intersect(*values)

    #operators
    def __or__(self, value):
        return self.add(value)
                
    def __and__(self, value):
        return self.intersect(value)
                
    def __sub__(self, value):
        return self.subtract(value)

    #primitives
    def rhomb(self, cell):
        origin = np.zeros(3)
        points = [x+y+z for x in (origin,cell[0]) for y in (origin,cell[1]) for z in (origin,cell[2])]
        s = "polyhedron(["
        for point in points:
            s += "[{0},{1},{2}],\n".format(*point)
        s += "],\n"
        faces = [[0,1,3,2],[0,4,5,1],[0,2,6,4],[5,4,6,7],[6,2,3,7],[3,1,5,7]]
        s += "["
        for face in faces:
            s += "[{3},{2},{1},{0}],\n".format(*face)
        s += "]);\n"
        return OpenScad(s)

    def bond(self, s1,s2,r=1.0):
        return OpenScad("bond([{0},{1},{2}],[{3},{4},{5}],r={6});\n".format(*s1,*s2,r))

    def sphere(self, r=1):
        return OpenScad("sphere(r={0});\n".format(r))

    def __str__(self):
        return self.string
    
def test():
    o = OpenScad()
    print(o.sphere(r=5).translate([1,2,3]))
    print(o.add(o.sphere(r=2), o.sphere(r=3)))
    print(o.sphere(r=2).add(o.sphere(r=3)).add(o.sphere(r=4))) #another way






def hook0(lattice):
    lattice.logger.info("Hook0: Preprocess.")
    for d in range(3):
        lattice.rep[d] += 2  #Extend the size,then cut off later.
    lattice.logger.info("Hook0: end.")

def hook2(lattice):
    scale=50
    roxy=0.07
    rbond=0.06
    lattice.logger.info("Hook2: Output water molecules in OpenSCAD format revised.")
    cellmat = lattice.repcell.mat
    rep = np.array(lattice.rep)
    trimbox    = lattice.cell.mat *np.array([(rep[i]-2) for i in range(3)])
    trimoffset = lattice.cell.mat[0]+lattice.cell.mat[1]+lattice.cell.mat[2]
    lattice.logger.info(lattice.repcell.mat)
    lattice.logger.info(lattice.cell.mat)

    margin = 0.2 # expansion relative to the cell size
    lower = (1.0 - margin) / rep
    upper = (rep - 1.0 + margin) / rep

    bonds = []
    for i,j in lattice.graph.edges(data=False):
        s1 =lattice.reppositions[i]
        s2 =lattice.reppositions[j]
        d = s2-s1
        d -= np.floor( d + 0.5 )
        lattice.logger.debug("Len {0}-{1}={2}".format(i,j,np.linalg.norm(d)))
        s2 = s1 + d
        if ( (lower[0] < s1[0] < upper[0] and lower[1] < s1[1] < upper[1] and lower[2] < s1[2] < upper[2] ) or
            (lower[0] < s2[0] < upper[0] and lower[1] < s2[1] < upper[1] and lower[2] < s2[2] < upper[2] ) ):
            bonds.append( (np.dot(s1,cellmat), np.dot(s2,cellmat)))

    nodes = []
    for s1 in lattice.reppositions:
        if lower[0] < s1[0] < upper[0] and lower[1] < s1[1] < upper[1] and lower[2] < s1[2] < upper[2]:
            nodes.append( np.dot(s1, cellmat) )

    o = OpenScad()
    objs = [o.sphere(r="Roxy").translate(node) for node in nodes] + [o.bond(s1,s2,r="Rbond") for s1,s2 in bonds]
    #operations
    ops = [bondfunc,
        o.defvar("$fn", 20),
        o.defvar("Roxy", roxy),
        o.defvar("Rbond", rbond),
        ( o.rhomb(trimbox).translate(trimoffset) & o.add(*objs) ).translate(-trimoffset).scale([scale,scale,scale])]
    s = o.encode(*ops)
    s = '//' + "\n//".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook2: end.")

hooks = {0:hook0, 2:hook2}



# coding: utf-8
"""
Povray format.
"""

from collections import defaultdict
import numpy as np

from genice import rigid


def Block(name, content):
    return " " + name + " { " + content + " } "

def Vector(v):
    return "<{0:.3f}, {1:.3f}, {2:.3f}> ".format(*v)

def Juxtapose(v):
    return ",".join(v)

def Atom(atomtype, pos):
    return Block( "sphere", Juxtapose( [Vector(pos), "R{0}".format(atomtype) ] ) + Block( "material", "MAT{0}".format(atomtype) ) ) + "\n"
    
def Bond(bondtype, pos1, pos2):
    return Block( "cylinder", Juxtapose( [Vector(pos1), Vector(pos2), "R{0}".format(bondtype)] ) + Block( "material", "MAT{0}".format(bondtype) ) ) + "\n"
    

def hook6(lattice):
    global nwateratoms
    lattice.logger.info("Hook6: Output water molecules in Povray format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    # prepare the reverse dict
    waters = defaultdict(dict)
    for atom in lattice.atoms:
        resno, resname, atomname, position, order = atom
        if "O" in atomname:
            waters[order]["O"] = position
        elif "H" in atomname:
            if "H0" not in waters[order]:
                waters[order]["H0"] = position
            else:
                waters[order]["H1"] = position
    s = '#include "default.inc"\n'
    s += "union {\n"
    for order, water in waters.items():
        O = water["O"]
        H0 = water["H0"]        
        H1 = water["H1"]        
        s += Atom("O",O)
        s += Atom("H",H0)
        s += Atom("H",H1)
        s += Bond("OH",O,H0)
        s += Bond("OH",O,H1)
    for i,j in lattice.spacegraph.edges(data=False):
        if i in waters and j in waters:  # edge may connect to the dopant
            O = waters[j]["O"]
            H0 = waters[i]["H0"]
            H1 = waters[i]["H1"]
            d0 = H0 - O
            d1 = H1 - O
            rr0 = np.dot(d0,d0)
            rr1 = np.dot(d1,d1)
            if rr0 < rr1 and rr0 < 0.245**2:
                s += Bond("HB",H0,O)
            if rr1 < rr0 and rr1 < 0.245**2:
                s += Bond("HB",H1,O)
    print(s, end="")
    nwateratoms = len(lattice.atoms)
    lattice.logger.info("Hook6: end.")


def hook7(lattice):
    global nwateratoms
    lattice.logger.info("Hook7: Output water molecules in Povray format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    cellmat = lattice.repcell.mat
    gatoms = lattice.atoms[nwateratoms:]
    s = ""
    H = []
    O  = ""
    for atom in gatoms:
        resno, resname, atomname, position, order = atom
        s += Atom(atomname,position)
    s = '//' + "\n//".join(lattice.doc) + "\n" + s
    s += "  translate " + Vector( -(cellmat[0,:]+cellmat[1,:]+cellmat[2,:])/2 ) + "\n}\n"
    print(s)
    lattice.logger.info("Hook7: end.")
    

hooks = {7:hook7, 6:hook6}



#!/usr/bin/env python3
# -*- python -*-

import os
import sys
import argparse  as ap
import logging
from genice.importer import safe_import
from genice import lattice
import random
import numpy as np

def getoptions():
    parser = ap.ArgumentParser(description='')
    parser.add_argument('--rep',  '-r', nargs = 3, type=int,   dest='rep',  default=[1,1,1],
                        help='Repeat the unit cell in x,y, and z directions. [1,1,1]')
    parser.add_argument('--dens', '-d', nargs = 1, type=float, dest='dens', default=(-1,),
                        help='Specify the ice density in g/cm3')
    parser.add_argument('--seed', '-s', nargs = 1, type=int,   dest='seed', default=(1000,),
                        help='Random seed [1000]')
    parser.add_argument('--format', '-f', nargs = 1,           dest='format', default=("gromacs",), metavar="gmeqdypoc",
                        help='Specify file format [g(romacs)|m(dview)|e(uler)|q(uaternion)|d(igraph)|y(aplot)|p(ython module)|o(penScad)|c(entersofmass)|r(elative com)] [gromacs]')
    parser.add_argument('--water', '-w', nargs = 1,           dest='water', default=("tip3p",), metavar="model",
                        help='Specify water model. (tip3p, tip4p, etc.) [tip3p]')
    parser.add_argument('--guest', '-g', nargs = 1,           dest='guests', metavar="D=empty", action="append", 
                        help='Specify guest(s) in the cage type. (D=empty, T=co2*0.5+me*0.3, etc.)')
    parser.add_argument('--Guest', '-G', nargs = 1,           dest='spot_guests', metavar="13=me", action="append", 
                        help='Specify guest in the specific cage. (13=me, 32=co2, etc.)')
    parser.add_argument('--Group', '-H', nargs = 1,           dest='groups', metavar="13=bu-:0", action="append", 
                        help='Specify the group. (-H 13=bu-:0, etc.)')
    parser.add_argument('--anion', '-a', nargs = 1,           dest='anions', metavar="3=Cl", action="append", 
                        help='Specify a monatomic anion that replaces a water molecule. (3=Cl, 39=F, etc.)')
    parser.add_argument('--cation', '-c', nargs = 1,           dest='cations', metavar="3=Na", action="append", 
                        help='Specify a monatomic cation that replaces a water molecule. (3=Na, 39=NH4, etc.)')
    parser.add_argument('--nodep', action='store_true', dest='nodep',
                        help='No depolarization.')
    parser.add_argument('--asis', action='store_true', dest='asis',
                        help='Assumes all given HB pairs to be fixed. No shuffle and no depolarization.')
    parser.add_argument('--debug', '-D', action='store_true', dest='debug',
                        help='Output debugging info.')
    parser.add_argument('--quiet', '-q', action='store_true', dest='quiet',
                        help='Do not output progress messages.')
    parser.add_argument('Type', nargs=1,
                       help='Crystal type (1c,1h,etc. See https://github.com/vitroid/GenIce for available ice structures.)')
    return parser.parse_args()


        

def main():
    # Module-loading paths
    # 1. Look for the modules in the current working directory
    sys.path.append(".")
    #prepare user's workarea
    home = os.path.expanduser("~")
    if os.path.exists(home+"/Library/Application Support"): #MacOS
        homegenice = home+"/Library/Application Support/GenIce"
    else:
        homegenice = os.path.expanduser(home + "/.genice") #Other unix
    try:
        os.makedirs(homegenice+"/formats")
        os.makedirs(homegenice+"/lattices")
        os.makedirs(homegenice+"/molecules")
    except:
        pass #just ignore when failed.
    # 2. Look for user's home.
    sys.path.append(homegenice)


    #Parse options
    options = getoptions()

    
    #Set verbosity level
    if options.debug:
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s %(levelname)s %(message)s")
    elif options.quiet:
        logging.basicConfig(level=logging.WARN,
                            format="%(levelname)s %(message)s")
    else:
        #normal
        logging.basicConfig(level=logging.INFO,
                            format="%(levelname)s %(message)s")
    logger = logging.getLogger()
    logger.debug("Debug mode.")
    logger.debug(options.Type)

    water_type   = options.water[0]
    guests       = options.guests
    lattice_type = options.Type[0]
    file_format  = options.format[0]
    seed         = options.seed[0]
    rep          = options.rep
    density      = options.dens[0]
    depolarize   = not options.nodep
    asis         = options.asis
    anions = dict()
    if options.anions is not None:
        logger.info(options.anions)
        for v in options.anions:
            key, value = v[0].split("=")
            anions[int(key)] = value
    cations = dict()
    if options.cations is not None:
        for v in options.cations:
            key, value = v[0].split("=")
            cations[int(key)] = value
    spot_guests = dict()
    if options.spot_guests is not None:
        for v in options.spot_guests:
            key, value = v[0].split("=")
            spot_guests[int(key)] = value
    groups = dict()
    if options.groups is not None:
        for v in options.groups:
            key, value = v[0].split("=")
            groups[int(key)] = value

    del options  # Dispose for safety.
    # Set random seeds
    random.seed(seed)
    np.random.seed(seed)
    
    logger.debug("Lattice: {0}".format(lattice_type))
    # Main part of the program is contained in th Formatter object. (See formats/)
    logger.debug("Format: {0}".format(file_format))
    formatter = safe_import("format", file_format)
    lat = lattice.Lattice(lattice_type,
                          density=density,
                          rep=rep,
                          depolarize=depolarize,
                          asis=asis,
                          cations=cations,
                          anions=anions,
                          spot_guests=spot_guests,
                          spot_groups=groups,
                          )
    # These arguments should also be in lattice, not in run()
    lat.format(water_type=water_type,
                guests=guests,
                formatter=formatter
                )
    
if __name__ == "__main__":
    main()



# coding: utf-8
"""
Test suite: ring direction statistics

Hollins, G. T. Configurational statistics and the dielectric constant of ice. Proc. Phys. Soc. 84, 1001–1016 (1964).
"""

# A directed cycle is expressed as an array of True and False.
# For example,
# --> --> --> : [True, True, False]
# --> <-- --> : [Frue, False, True]
#They are isomorphic if they can be identical by rotations and inversions.

# 


import itertools as it
from countrings import countrings_nx as cr
import sys
from collections import defaultdict
import networkx as nx
from math import log
import fractions

def isomorphs(a):
    """
    return the symmetry number of cycle a
    """
    iso = set()
    aa = a + a
    for i in range(len(a)):
        part = tuple(aa[i:i+len(a)])
        iso.add(part)
    e = [not x for x in a]
    ee = e + e
    for i in range(len(a)):
        part = tuple(ee[i:i+len(a)])
        iso.add(part)
    return iso


def code(ori):
    """
    from an array of True/False to an integer.
    """
    s = 0
    for i in range(len(ori)):
        if ori[i]:
            s += 1<<i
    return s


def encode(ori):
    m = 9999
    for i in isomorphs(ori):
        c = code(i)
        if c < m:
            m = c
    return m

    
def symmetry(a):
    return len(isomorphs(a))


def freedom(a):
    """
    return possible number of molecular placements
    """
    n = 1
    for i in range(len(a)):
        if a[i-1] is a[i]:
            n *= 2
    return n


def ideal_frequency(a):
    N = len(a)
    return symmetry(a) * freedom(a)


def contains(a,b):
    """
    return True if one of the isomorphs of b is included in the set a.
    """
    return len(a & isomorphs(b)) > 0
    

def ideal_frequencies(N):
    freq = dict()
    for a in it.product((False, True), repeat=N):
        c = encode(a)
        if c not in freq:
            freq[c] = ideal_frequency(a)
    return freq


def probabilities(N):
    freq = ideal_frequencies(N)
    sum = 0
    for v in freq:
        sum += freq[v]
    prob = dict()
    for v in freq:
        prob[v] = fractions.Fraction(freq[v], sum)
    return prob


def orientations(members, digraph):
    return [digraph.has_edge(members[i-1], members[i]) for i in range(len(members))]


def hook4(lattice):
    lattice.logger.info("Hook4: Ring test.")
    graph = nx.Graph(lattice.spacegraph) #undirected
    stat = dict()
    prob = defaultdict(int)
    for n in 3,4,5,6,7,8:
        prob[n] = probabilities(n)
        stat[n] = defaultdict(int)
    for ring in cr.CountRings(graph).rings_iter(8):
        ori = orientations(ring, lattice.spacegraph)
        c   = encode(ori)
        n   = len(ring)
        stat[n][c] += 1
    #size code code(binary) Approx. Stat.
    for n in 3,4,5,6,7,8:
        fmtstr = "{{0}} {{1}} {{1:0{0}b}} {{2}} {{3:.5f}} {{4:.5f}}".format(n)
        denom = 0
        for c in stat[n]:
            denom += stat[n][c]
        if denom > 0:
            dKL = 0.0
            for c in prob[n]:
                print(fmtstr.format(n,c,prob[n][c], float(prob[n][c]), stat[n][c]/denom))
                q = stat[n][c]/denom
                p = prob[n][c]
                if q > 0.0:
                    dKL += q*(log(q) - log(p))
            dKL /= log(2.0)  #bit
            print("{1} {0} dKL[{0}-ring]".format(n,dKL))
            lattice.logger.info("  dKL[{0}-ring]={1}".format(n,dKL))
    #print(score)
    lattice.logger.info("Hook4: end.")


hooks = {4:hook4}



#!/usr/bin/env python3


import numpy as np
import logging
from math import pi, sin, cos

#a sample for general alkyl group with methyls
# tree | branch: [root,branch1,branch2,branch3]
# or             [root,branch1,branch2]
# or             [root,branch1]
# or             [leaf]
# or             leaf
# 3-methylbutyl : backbone=["Ma",["Mb",["Mc",["Md","Me"]]]]]
#v1 and v2 must be given as a unit vector.
def alkyl(direction, destination, tree, dest=None):
    """
    put a normal-alkyl group rooted at root toward cpos.
    """
    logger = logging.getLogger()
    logger.debug("  {0}".format(tree))
    # logger.info("  Put butyl at {0}".format(molname))

    if type(tree) is not list:
        tree = [tree]

    # v2 is a vector from direction to the destination
    v2 = destination - direction
    v2 /= np.linalg.norm(v2)
    v2d = np.dot(direction, v2)
    while v2d > 0.999:
        #They are inline. It is not safe to determine the orientation.
        v2 = np.random.random(3)
        v2 /= np.linalg.norm(v2)
        v2d = np.dot(direction, v2)

    v2 -= v2d * direction
    v2 /= np.linalg.norm(v2)

    #v1 is the pivot
    v1 = direction
    
    v3 = np.cross(v1, v2)     # the thild unit vector

    c = cos(120 * pi / 180)
    s = sin(120 * pi / 180)
    v4 = v2*c + v3*s   # a branch vector
    v5 = v2*c - v3*s   # another branch vector
    logger.debug("  {0} -0.5".format(np.dot(v2, v4)))
    logger.debug("  {0} -0.5".format(np.dot(v4, v5)))
    logger.debug("  {0} -0.5".format(np.dot(v5, v2)))

    c = cos(109.5 * pi / 180)
    s = sin(109.5 * pi / 180)
    v2 = -v1*c + v2*s
    v4 = -v1*c + v4*s
    v5 = -v1*c + v5*s
    logger.debug("  {0} 1/3".format(np.dot(v1, v2)))
    logger.debug("  {0} 1/3".format(np.dot(v1, v4)))
    logger.debug("  {0} 1/3".format(np.dot(v1, v5)))
    logger.debug("  {0} 1".format(np.dot(v2, v2)))
    logger.debug("  {0} 1".format(np.dot(v4, v4)))
    logger.debug("  {0} 1".format(np.dot(v5, v5)))

    #assert False
    atomname = tree[0]
    atoms = [(atomname, np.zeros(3))]
    for vec,topo in zip([v2,v4,v5], tree[1:]):
        atoms += alkyl(vec, destination - v1, topo)
    #untranslation
    atoms = [(atom[0], atom[1] + v1) for atom in atoms]
    
    return atoms


def test():
    logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s %(levelname)s %(message)s")
    direction = np.array([1.0, 0.0, 0.0])    #must be a unit vector
    destination = np.array([10.0, 10.0, 0.0])  #All the branches direct to this point.
    atoms = alkyl(direction, destination, tree=["Ma",["Mb", "Mf", ["Mc", "Me", ["Md", ["A", ["B", ["C", ["D"]]]]]]]])
    # in yaplot format
    print("t 0 0 0 +")
    print("t",destination[0], destination[1], destination[2], "@")
    for atom in atoms:
        name, pos = atom
        print("t",pos[0], pos[1], pos[2], name)


if __name__ == "__main__":
    test()



#!/usr/bin/env python3

from setuptools import setup
import os
import codecs
import re

#Copied from wheel package
here = os.path.abspath(os.path.dirname(__file__))
#README = codecs.open(os.path.join(here, 'README.txt'), encoding='utf8').read()
#CHANGES = codecs.open(os.path.join(here, 'CHANGES.txt'), encoding='utf8').read()

with codecs.open(os.path.join(os.path.dirname(__file__), 'genice', '__init__.py'),
                 encoding='utf8') as version_file:
    metadata = dict(re.findall(r"""__([a-z]+)__ = "([^"]+)""", version_file.read()))

setup(name='GenIce',
      version=metadata['version'],
      description='A Swiss army knife to generate hydrogen-disordered ice structures.',
      #long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        ],
      author='Masakazu Matsumoto',
      author_email='vitroid@gmail.com',
      url='https://github.com/vitroid/GenIce/',
      keywords=['genice',],
      license='MIT',
      packages=['genice',
                'genice.molecules',
                'genice.lattices',
                'genice.formats',
                ],
      # install_requires=['numpy', 'networkx', 'countrings>=0.1.4', 'svgwrite', 'vpython'],
      install_requires=['numpy', 'networkx', 'countrings>=0.1.4', 'svgwrite', 'cif2ice>=0.2'],
      entry_points = {
              'console_scripts': [
                  'genice = genice.__main__:main'
              ]
          }
#      entry_points="""
#      [console_scripts]
#      genice = genice.__main__:main
#      """
      )




# -*- coding:utf-8 -*-

"""
Load a gromacs file as a lattice.

Usage: genice gromacs[filename]            Regards Ow and Hw to be of a water molecule.
       genice gromacs[filename:Ow]         Specify the name of O atom (and ignore H positions)
       genice gromacs[filename:Ow:Hw[12]]  Specify the names of O and H atoms. (Regexp is accepted.)

"""


import numpy as np
import re
import logging
import genice.pairlist as pl

def argparser(arg):
    global waters, cell, celltype, coord, density, pairs
    logger = logging.getLogger()
    args = arg.split(":")
    assert 0 < len(args) <= 3, __doc__
    if len(args) == 1:
        O = "Ow"
        H = "Hw"
    elif len(args) == 2:
        O = args[1]
        H = None
    else:
        O = args[1]
        H = args[2]
    filename = args[0]

    file = open(filename)
    file.readline()
    natom = int(file.readline())
    hatoms = []
    waters = []
    for i in range(natom):
        line = file.readline()
        # resid = int(line[0:5])
        # resna = line[5:10]
        atomname = line[10:15].replace(' ', '')
        # atomid = int(line[15:20])
        pos = np.array([float(x) for x in line[20:].split()[:3]]) #drop velocity
        if atomname == O:
            waters.append(pos)
        elif H is not None and re.fullmatch(H, atomname):
            hatoms.append(pos)
        else:
            logger.info("Skip {0}".format(atomname))
    c = [float(x) for x in file.readline().split()]
    if len(c) == 3:
        cell = np.array([[c[0],0.,0.],
                         [0.,c[1],0.],
                         [0.,0.,c[2]]])
    else:
        cell = np.array([[c[0],c[3],c[4]],
                         [c[5],c[1],c[6]],
                         [c[7],c[8],c[2]]])
    celltype = 'triclinic'
    coord = 'absolute'
    density = len(waters) / (np.linalg.det(cell)*1e-21) * 18 / 6.022e23

    if len(hatoms) > 0:
        celli = np.linalg.inv(cell)
        # relative coord
        rh = [np.dot(x, celli) for x in hatoms]
        ro = [np.dot(x, celli) for x in waters]
        grid = pl.determine_grid(cell, 0.245)
        pairs0 = pl.pairlist_fine_hetero(ro, rh, 0.245, cell, grid, distance=False)
        # remove intramolecular OHs
        pairs = []
        for o,h in pairs0:
            if h == o*2 or h == o*2+1:
                # adjust oxygen positions
                dh = rh[h] - ro[o]
                dh -= np.floor(dh + 0.5)
                waters[o] += np.dot(dh, cell)*1./16.
            else:
                # register a new intermolecular pair
                pairs.append((h//2, o))
        logger.debug("# of pairs: {0} {1}".format(len(pairs),len(waters)))
        
    
    

# default. Do nothing (leading to an error).




# coding: utf-8
"""
[1] J. L. F. Abascal, E. Sanz, R. G. Fernández, and C. Vega, A potential model for the study of ices and amorphous water: TIP4P/Ice, J. Chem. Phys. 122 (2005) 234511.
"""

import numpy as np
from math import pi,sin,cos

L1 = 0.9572 / 10
L2 = 0.1577 / 10
theta=104.52 * pi/180


hy = L1*sin(theta/2)
hz = L1*cos(theta/2)
mz = L2

sites = np.array([[0.0, 0.0, 0.0],
                  [0.0, hy,  hz],
                  [0.0,-hy,  hz],
                  [0.0, 0.0, mz]])
sites -= (sites[1]+sites[2]+sites[0]*16)/18

                  
atoms = ["O","H","H","."]
labels = ["OW","HW1","HW2","MW"]
name = "ICE"





# coding: utf-8
import math
import numpy as np


oh = 0.09572 #nm
om = 0.07    #nm
hangle = 104.52 * math.pi / 180 / 2
mangle = 109.47 * math.pi / 180 / 2
mass=18
ohz = oh * math.cos(hangle)
ohy = oh * math.sin(hangle)
omz =-om * math.cos(mangle)
omx = om * math.sin(mangle)
oz  = -ohz*2/mass


sites = np.array([[0, 0,oz],
                  [0, ohy,ohz+oz],
                  [0,-ohy,ohz+oz],
                  [ omx,0,omz+oz],
                  [-omx,0,omz+oz],
                  ]) # nm, OHHMM

atoms = ["O","H","H",".","."]
labels = ["OW","HW1","HW2","MW1","MW2"]
name = "SOL"



# coding: utf-8
"""
Yaplot format.
defined in https://github.com/vitroid/Yaplot
"""

from collections import defaultdict
import numpy as np
import networkx as nx

from genice import rigid
from genice import yaplotlib as yp
from countrings import countrings_nx as cr




def face(center, rpos):
    pos = rpos + center
    n = rpos.shape[0]
    s = yp.Color(n)
    s += yp.Layer(n)
    s += yp.Polygon(pos)
    return s



def hook4(lattice):
    lattice.logger.info("Hook4: A. Output depolarization process in Yaplot format.")
    print(lattice.yapresult, end="")
    lattice.logger.info("Hook4: B. Polyhedral expression.")
    # copied from svg_poly
    graph = nx.Graph(lattice.spacegraph) #undirected
    cellmat = lattice.repcell.mat
    for ring in cr.CountRings(graph).rings_iter(8):
        deltas = np.zeros((len(ring),3))
        d2 = np.zeros(3)
        for k,i in enumerate(ring):
            d = lattice.reppositions[i] - lattice.reppositions[ring[0]]
            d -= np.floor(d+0.5)
            deltas[k] = d
            dd = lattice.reppositions[ring[k]] - lattice.reppositions[ring[k-1]]
            dd -= np.floor(dd+0.5)
            d2 += dd
        # d2 must be zeros
        if np.all(np.absolute(d2) < 1e-5):
            comofs = np.sum(deltas, axis=0) / len(ring)
            deltas -= comofs
            com = lattice.reppositions[ring[0]] + comofs
            com -= np.floor(com)
            # rel to abs
            com    = np.dot(com,    cellmat)
            deltas = np.dot(deltas, cellmat)
            print(face(com,deltas), end="")
    print()
    lattice.logger.info("Hook4: end.")


def hook6(lattice):
    global nwateratoms
    lattice.logger.info("Hook6: Output water molecules in Yaplot format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    # prepare the reverse dict
    waters = defaultdict(dict)
    for atom in lattice.atoms:
        resno, resname, atomname, position, order = atom
        if "O" in atomname:
            waters[order]["O"] = position
        elif "H" in atomname:
            if "H0" not in waters[order]:
                waters[order]["H0"] = position
            else:
                waters[order]["H1"] = position
    s = ""
    s += yp.Color(3)
    for order, water in waters.items():
        O = water["O"]        
        H0 = water["H0"]        
        H1 = water["H1"]        
        s += yp.Layer(4)
        s += yp.Color(3)
        s += yp.Size(0.03)
        s += yp.Circle(O)
        s += yp.Line(O,H0)
        s += yp.Line(O,H1)
        s += yp.Size(0.01)
        s += yp.Circle(H0)
        s += yp.Circle(H1)
        s += yp.Color(2)
        s += yp.Layer(1)
        s += yp.Text(O, "{0}".format(order))
    s += yp.Layer(3)
    s += yp.Color(4)
    s += yp.ArrowType(1)
    s += yp.Size(0.03)
    for i,j in lattice.spacegraph.edges(data=False):
        if i in waters and j in waters:  # edge may connect to the dopant
            O = waters[j]["O"]
            H0 = waters[i]["H0"]
            H1 = waters[i]["H1"]
            d0 = H0 - O
            d1 = H1 - O
            rr0 = np.dot(d0,d0)
            rr1 = np.dot(d1,d1)
            if rr0 < rr1 and rr0 < 0.27**2:
                s += yp.Arrow(H0,O)
            if rr1 < rr0 and rr1 < 0.245**2:
                s += yp.Arrow(H1,O)
    print(s, end="")
    nwateratoms = len(lattice.atoms)
    lattice.logger.info("Hook6: end.")


def hook7(lattice):
    global nwateratoms
    lattice.logger.info("Hook7: Output water molecules in Yaplot format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    gatoms = lattice.atoms[nwateratoms:]
    palettes = dict()
    s = ""
    s += yp.Layer(4)
    s += yp.ArrowType(1)
    H = []
    O  = ""
    for atom in gatoms:
        resno, resname, atomname, position, order = atom
        if atomname in palettes:
            pal = palettes[atomname]
        else:
            pal = 4 + len(palettes)
            palettes[atomname] = pal
        s += yp.Color(pal)
        s += yp.Size(0.04)
        s += yp.Circle(position)
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    print(s)
    lattice.logger.info("Hook7: end.")
    

hooks = {7:hook7, 6:hook6, 4:hook4}



import re
import importlib
import logging

#
# Accept options parenthesized after the plugin name.
#


def audit_name(name):
    """
    Audit the mol name to avoid the access to external files
    """
    return re.match('^[A-Za-z0-9-_]+$', name) is not None



def safe_import(category, name):
    logger = logging.getLogger()
    assert category in ("lattice", "format", "molecule")
    # name may contain arguments
    left = name.find("[")
    right = name.rfind("]")
    arg = ""
    if left < right:
        arg = name[left+1:right]
        name = name[:left]
    assert audit_name(name), "Dubious {0} name: {1}".format(category, name)
    module = None
    try:
        module     = importlib.import_module(category+"s."+name) #at ~/.genice
    except ImportError as e:
        pass
    if module is None:
        fullname = "genice."+category+"s."+name
        logger.debug("Load module: {0}".format(fullname))
        module     = importlib.import_module(fullname)
    logger.info("Load {0} module {1} arguments [{2}]".format(category, name, arg))
    if arg != "":
        if "argparser" in module.__dict__:
            module.argparser(arg)
        else:
            logger.info("Arguments are given but the module does not accept them.")
    elif "usage" in module.__dict__:
        module.usage()
    return module



# coding: utf-8
"""
Note: Due to the technical limitation in the GenIce algorithm, the minimum lattice size is larger than the crystallographic unit cell size.
"""

from math import sin, pi, cos
import logging

def usage():
    logger = logging.getLogger()
    logger.info("** prism module **")
    logger.info("prism module accepts two arguments.")
    logger.info("prism[sides,rows]")
    logger.info("By default, sides are 6 and rows are 10.")
    logger.info("prism[5] prepares pentagonal prism of 10 rows.")
    logger.info("prism[8,5] prepares octagonal prism of 5 rows.")
    logger.info("------------------")
    


def argparser(arg):
    global sides, rows, bondlen, density, cell, celltype, waters, coord
    a = [int(x) for x in arg.split(",")]
    if len(a)>0:
        sides = a[0]
    if len(a)>1:
        rows = a[1]
    L = 2.75
    bondlen = 3
    R = L/2/sin(pi/sides)
    celltype = "rect"
    cell = [L*20, L*20, L*rows]
    density = sides*rows / (L**3 * 400 * rows) * 18 / 6.022e23 * 1e24

    waters = []
    for j in range(rows):
        for i in range(sides):
            x = R * cos(i*pi*2/sides)
            y = R * sin(i*pi*2/sides)
            z = j * L
            waters.append([x,y,z])

    coord = "absolute"


# default.
argparser("6,10")



#!/usr/bin/env python3

#FAU Decoration of a 4-network
#読みこんだAR3Rの座標を、FAU構造における多面体vertexの位置とみなし、
#それらを連結するネットワークを六角柱で修飾して大きなネットワークを作る。

from math import acos, pi, sin, cos
from collections import defaultdict
import numpy as np
import logging
import re


def tune_angles(sixvecs, pivot):
    """
    Find the best origin of angles to make sum cos(3 th) largest
    """
    sixangles = []
    for i in range(len(sixvecs)):
        vec = sixvecs[i]
        cosine = np.dot(sixvecs[0],vec)
        if cosine > 1.0:
            cosine = 1.0
        angle = acos(cosine)
        sine   = np.cross(sixvecs[0], vec)
        if np.dot(sine, pivot) < 0:
            angle = -angle
        sixangles.append(angle)
    offset = 0
    while True:
        sum = 0.0
        dsum = 0.0
        for a in sixangles:
            sum += cos((a+offset)*6)
            dsum += -sin((a+offset)*6)
        doffset = dsum / 20.0
        if abs(doffset) < 1e-6:
            return offset
        offset += doffset
    

    
class decorate():
    def __init__(self, atoms, cell, pairs, Ncyl):
        """
        Ncyl is the number of cylinders to be inserted (>0)
        """
        #make netghbor list
        nei = defaultdict(set)
        for i,j in pairs:
            nei[i].add(j)
            nei[j].add(i)
        self.nei = nei
        self.atoms = atoms
        self.cell  = cell
        self.Ncyl  = Ncyl
        self.vertices = []
        self.fixedEdges = []
        for pair in pairs:
            self.one(pair)

    def one(self, pair):
        i,j = pair
        dij = self.atoms[j] - self.atoms[i]
        dij -= np.floor(dij + 0.5)
        dij = np.dot(dij, self.cell)
        scale = np.linalg.norm(dij) 
        dij /= scale
        sixpairs = []
        for k in self.nei[i]:
            if k != j:
                sixpairs.append((k,i))
        for k in self.nei[j]:
            if k != i:
                sixpairs.append((k,j))
        #Regularize the dihedral angles
        #to point them 6-fold directions.
        #by adding an offset
        sixvecs = []
        for j,k in sixpairs:
            vec = self.atoms[j] - self.atoms[k]
            vec -= np.floor(vec + 0.5)
            vec = np.dot(vec, self.cell)
            #orthogonalize
            shadow = np.dot(dij, vec)
            vec -= shadow*dij
            vec /= np.linalg.norm(vec)
            #print(np.dot(vec,dij))
            sixvecs.append(vec)
        offset = tune_angles(sixvecs, dij)
        offset += pi/6 #30 degree
        x = sixvecs[0].copy()
        z = dij
        y = np.cross(z,x)
        for j in range(6):
            a = j*pi*2/6 + offset
            sixvecs[j] = x*cos(a) + y*sin(a)
        #determine r
        #assume edge length is 1
        #the radius of the outer sphere of the polyhed is sqrt(3/2)
        L = (3/2)**0.5 * 2 + self.Ncyl
        r = 1/L     #edge len = radius of cyl
        rp = (3/2)**0.5 / L  # = radius of polyhed
        #
        a = np.dot(self.atoms[i],self.cell)
        s = ""
        for j in range(0, self.Ncyl+1):
            vec0 = dij*(rp + j*r)*scale + a
            for vec in sixvecs:
                rpos = vec0 + vec*r*scale
                pos = np.dot(rpos, np.linalg.inv(self.cell))
                self.vertices.append(pos)
            first = len(self.vertices)-6
            if j % 2 == 0:
                for k in range(5):
                    self.fixedEdges.append((first+k, first+k+1))
                self.fixedEdges.append((first+5,first))
            else:
                for k in range(5):
                    self.fixedEdges.append((first+k+1, first+k))
                self.fixedEdges.append((first,first+5))
            if j > 0:
                for k in range(6):
                    if k % 2 == 0:
                        self.fixedEdges.append((first+k, first+k-6))
                    else:
                        self.fixedEdges.append((first+k-6, first+k))


logger = logging.getLogger()
from genice.lattices import ice1c # base topology
cell1c = np.diag(np.fromstring(ice1c.cell, sep=" "))
waters1c = np.fromstring(ice1c.waters, sep=" ")
waters1c = waters1c.reshape((waters1c.shape[0]//3,3))
pairs1c = np.fromstring(ice1c.pairs, sep=" ", dtype=int)
pairs1c = pairs1c.reshape((pairs1c.shape[0]//2,2))


def argparser(arg):
    global Ncyl, coord, celltype, cell, waters, fixed
    assert re.match("^[0-9]+$", arg) is not None, "Argument must be an integer."
    Ncyl = int(arg)
    logger.info("Superlattice {0}xFAU".format(Ncyl))
    dec = decorate(waters1c, cell1c, pairs1c, Ncyl)
    coord='relative'
    celltype='rect'
    cell = "{0} {1} {2}".format(dec.cell[0,0],dec.cell[1,1],dec.cell[2,2])
    waters = dec.vertices
    fixed = dec.fixedEdges

# default.
argparser("1")
